; ----------------------------------------------------------------------------
; Name:     Multi system BIOS MC68000/MC68020
; Author:   Steven
; Date:     04-07-1991 / 04-02-1993 / 09-01-2022
; Revision: 22-08-2021 Reloceerbaar tbv testen in RAM
;           12-11-2021 Timer/interrupts herschreven/Easy68K traps
;           02-01-2022 S-Record lader
;           20-02-2022 Diverse voorbereidingen op FUZIX
;           09-06-2023 NANO68K ondersteuning
;           02-12-2023 EC020-SBC ondersteuning
;           21-08-2024 Thunderbird support
; ----------------------------------------------------------------------------

;To do:
;12-02-2022 Statusbytes omzetten naar statusbits?
;04-02-2022 TESTEN: S-load tbv FUZIX (SU)+queue benchmark
;           TESTEN: TRAP-12, Trap-15#24,FuzixIRQ
;02-04-2024 Sources afsplitsen (bv biosmenu.asm)
;17-04-2024 Voor 2x cpu opgeblazen (cache kapot?) Standaard uitzetten 
;11-05-2024 Nieuwe (3e) print geassembleerd

;*** NEW ***
;28-05-2024 Backspace werkt niet meer in regeleditor (Dhrystone)
;11-06-2024 Program terminated: overschrijft start/RUN-adres
;12-06-2024 Trap 15 optimaliseren door gebruik juiste registers
;15-09-2024 BIOSinRAM automatisch berekenen

;21-12-2024 (F)>SIEVE crasht soms omdat (P?) de stack niet goed
;terugzet? Waarschijnlijk blijft de kleine SU stack staan, wat te
;weinig is voor sieve. Eerst uitzoeken wanneer het niet stabiel is


; ============================================================================
; This BIOS can handle four different 68K boards: 
System68K_Classic     equ   0  ;07-1991 Classic:      MC68HC000, 128kB RAM, 6 250 UART, 6522 VIA
System68K_Nano        equ   0  ;07-2023 68k-Nano:     MC68HC000,   1MB RAM, 16550 UART
SystemEC020SBC        equ   0  ;10-2023 ByteBlaze020: MC68020,   512kB RAM, 16550 UART
System68K_Thunderbird equ   1  ;08-2024 Thunderburd:  MC68HC000,   1MB RAM, 16550 UART

BIOSinRAM             equ   1  ;BIOS in RAM=1, BIOS in ROM=0


; ============================================================================
lf          equ           10              ;Linefeed[10]
cr          equ           13              ;Cariage return[13]
bs          equ            8              ;Backspace[8]            
Xon         equ           17
Xoff        equ           19
esc         equ           27
del         equ          127              ;Toets backspace = #127

;X-modem definities
SOH         equ          $01              ;SOH[1]
EOT         equ          $04              ;EOT[1]
ACK         equ          $06              ;ACK
NAK         equ          $15              ;NOT ack
CAN         equ          $18              ;CAN = cancel
XModemPoly  equ          $1021            ;Poly voor XModem-CRC16

RecBufLen   equ          128              ;Lengte ontvang buffer. Moet 2^n zijn!
Bovengrens  equ          100              ;Bij xx tekens in de buffer een Xoff zenden
Ondergrens  equ           64              ;Bij xx tekens in de buffer een Xon zenden


;Acia commando: /16, 8 data, 1 stop, geen pariteit, ontvangint aan, RTS laag.
AciaReset   equ          $43              ;Reset commando acia
AciaZonder  equ          $55              ;Acia commando zonder interrupts
AciaTrOff   equ          $95              ;Zend interrupt uit
AciaTrOn    equ          $B5              ;Zend interrupt aan

            
; ============================================================================
; System definements
            if System68K_Classic System68K_Classic
RAM_offset  equ         $010000          ;Start adres RAM
VAR_offset  equ         RAM_offset       ;Start adres variabelen
SStack      equ	      $010400		     ;Ruimte voor de supervisor stack (1kb)
Program     equ	      $010400		     ;Laad/run adres van programma's
RAM_end     equ         $02FFFF          ;Einde van 128 kB RAM
ROM_offset  equ         $000000          ;ROM op $0000
UART16550   equ         0                ;No UART 16550 on classic
            endif
            
            if System68K_Nano
RAM_offset  equ         $C00000          ;Start adres RAM
VAR_offset  equ         RAM_offset       ;Start adres variabelen
SStack      equ	      $C00400		     ;Ruimte voor de supervisor stack (1kb)
Program     equ	      $C00400		     ;Laad/run adres van programma's
RAM_end     equ         $CFFFFF          ;Einde van 1 MB RAM
ROM_offset  equ         $000000          ;ROM op $0000
UART16550   equ         $A00000          ;UART 16550 start address
RelCPUspeed equ         39               ;Calculating CPU speed by writing 16 char
            endif
            
            if SystemEC020SBC
RAM_offset  equ         $080000          ;Start adres RAM
VAR_offset  equ         RAM_offset       ;Start adres variabelen
SStack      equ	      $080400		     ;Ruimte voor de supervisor stack (1kb)
Program     equ	      $080400		     ;Laad/run adres van programma's
RAM_end     equ         $100000          ;Einde van 512kB RAM
ROM_offset  equ         $000000          ;ROM op $0000
UART16550   equ         $180000          ;UART 16550 start address
RelCPUspeed equ         39               ;Calculating CPU speed by writing 16 char
            endif

            if System68K_Thunderbird
RAM_offset  equ         $000000          ;Start adres RAM
VAR_offset  equ         RAM_offset+$400  ;Start adres variabelen na vectoren
SStack      equ	      $000800		     ;Ruimte voor de supervisor stack (1kb)
Program     equ	      $000800		     ;Laad/run adres van programma's
RAM_end     equ         $0FFFFF          ;Einde van 1MB RAM
UART16550   equ         $100000          ;UART 16550 start address

            if BIOSinRAM
ROM_offset  equ         $000000          ;Loading in RAM, always from 0x0000
            else
ROM_offset  equ         $200000          ;EEPROM W27C512 (2x64kB)
            endif;(BIOSinRAM)

EDI_offset  equ         $300000          ;EDI/CF-offset
RelCPUspeed equ         11               ;Calculating CPU speed by writing 16 char
            endif;(System68K_Thunderbird)

            
;--- UART 6850 ---------------------------------------------------------------
AciaCmd     equ         $030001           ;Commandoregister 6850 acia
AciaData    equ	      $030003           ;Bufferregister acia

MagicNumber equ         $55AAFF00         ;Controlegetal RAM-geheugen
DefaultClck equ         8000000           ;Default cpu klok 8 MHz
Timer1Freq  equ         100               ;100Hz (10ms) voor timer-1
Timer2Ref   equ         10                ;10 ms default one-shot timer-2


; --- VIA/6522 ---------------------------------------------------------------
ViaPB       equ         $030000		      ;Poort B
ViaPA       equ         $030002		      ;Poort A
ViaDDRB     equ         $030004		      ;Directionregister poort B
ViaDDRA     equ         $030006		      ;Directionregister poort A
ViaT1CL		equ	      $030008	         ;Timer 1 counterreg low
ViaT1CH		equ	      $03000A		      ;Timer 1 counterreg high
ViaT1LL		equ	      $03000C		      ;Timer 1 latch low
ViaT1LH		equ	      $03000E		      ;Timer 1 latch high
ViaT2CL		equ	      $030010		      ;Timer 2 counterreg low
ViaT2CH		equ	      $030012		      ;Timer 2 counterreg high
ViaSR		   equ		   $030014		      ;Shift register
ViaACR		equ	      $030016
ViaPCR		equ	      $030018           ;Peripheral control register
ViaIFR		equ	      $03001A		      ;Interrupt Flag Register
ViaIER		equ	      $03001C		      ;Interrupt Enable Register
ViaORA		equ	      $03001E


; --- Poorten PA/PB (bitnummers) ---------------------------------------------
PA0_SDA     equ         $00            ;Poort PA(0) I2C-data
PA1_SCL     equ         $01            ;Poort PA(1) I2C-klok
PA2_1wire   equ         $02            ;Poort PA(2) 1-wire (data/klok)

PB2_LED     equ         $02            ;Poort PB(2) LED (0=aan, 1=uit)
IFR_CA1     equ         $01            ;IFR CA1 is bit 1 is CA1 (1 sec tick RTC)
IFR_TIMER1  equ         $06            ;IFR Timer-1 is bit 6 (100Hz)
IFR_IRQ     equ         $07            ;IFT IRQ is bit 7 

IER_CA1     equ         1              ;IER CA1 (bit 1) 
IER_tmr1    equ         6              ;IER timer-1 interrupt (bit 6)
IER_enable  equ         7              ;IER enable interrupts (bit 7)


; ---Status 1&2 bits ---------------------------------------------------------
St1FuzixCall   equ         0           ;Jump to Fuzix
St1SRecError   equ         1           ;Reading SREC still OK


            section      Vectors,data
;Without repeating the "org" twice, the first vectors are maped to location 0x0000???
            org          ROM_offset    ;Start 68000 vectors
            org          ROM_offset    ;Start 68000 vectors
            
RomVector_00:dc.l        SStack        ;Stackpointer (points to RAM)
            dc.l         Start         ;Programcounter
            dc.l         BusErr        ;Bus error
            dc.l         AddrErr       ;Address error
            dc.l         IllErr        ;Illegale instructie
            dc.l         DivErr        ;Division by zero
            dc.l         ChkErr        ;CHK Grens-test
            dc.l         OvErr         ;TRAPV Overflow
            dc.l         PrivErr       ;Privilege fout

            dc.l         TraceErr      ;Enkele stap
            dc.l         IllErr        ;Line A emulator
            dc.l         IllErr        ;Line F emulator


            org          ROM_offset+$000060
            dc.l         SpurErr       ;Spurious error
            dc.l         ISR16550      ;Level-1 interrupt vector UART 16550

            org          ROM_offset+$000070       ;Level 4 interrupt autovector
            dc.l         AciaInt       ;Acia 6850-interrupt routine

            org          ROM_offset+$000080
            dc.l         Trap0EndProgram;Trap #0: einde userprogramma
            dc.l         Trap1Super    ;Trap #1: naar super-status

            org          ROM_offset+$0000B0       ;Trap #12: FUZIX-aanroep
            dc.l         Trap12Fuzix

            org          ROM_offset+$0000BC       ;Trap #15: Easy68K BIOS-routines
            dc.l         Trap15BIOS


            org          ROM_offset+$000100       ;Leesbare ODD/EVEN-tekst
            if SystemEC020SBC
             machine 68020              ;Assembler code voor MC68020
             dc.b         'MC68EC020-SBC';MC68EC020 heeft 1 EEPROM
            elseif            
             dc.b         'EOVDEDN--EEEEPPRROOMM',0;MC68000/68K-nano hebben ODD/EVEN-eeprom
            endif

   
; --- VARIABELEN -------------------------------------------------------------
; Vanaf 0x1.0000 t/m 0x1.0400 (1KB) voor UART buffers, variabelen en SuperStack 
; LET OP: ZONDER LABEL "ram" klopt de S-file niet
; LET OP: Achter RS.x MOET een getal staan; anders valt alles samen!
;x           section      Var,data
;ram         org          $010000       ;Start (1e 64 K) RAM
            rsset    VAR_offset   ;Variabelen in RAM

MaxMem      rs.l     1           ;Het eerste adres dat geen RAM meer is(L)
TimerTicks  rs.l     1           ;Tijd van de dag in 1/100 seconde obv Timer-1(L)
TimeSec     rs.l     1           ;Tijd van de dag in secondes obv RTC(L)
CPUclock    rs.l     1           ;Gemeten of default CPU clock (Hz)(L)
LeapTime    rs.l     1           ;Leap time(L)
Ident       rs.l     1           ;Hier staat na een koude-start een teken(L)

RecBuf      rs.b     RecBufLen   ;Ontvang buffer (128 bytes)

RecLeesPtr  rs.w     1           ;Ontvang-lees wijzer(W)
RecAantal   rs.w     1           ;Aantal tekens in ontvangbuffer(W)
;TrLeesPtr   rs.w     1          ;Zend-lees wijzer(W)
;TrAantal    rs.w     1          ;Aantal tekens in zendbuffer(W)

ProtocolOn  rs.b     1           ;Xon/Xoff protocol aan/uit(B)
XonXoff     rs.b     1           ;AciaCmd; bevat AciaTrOn of AciaTrOff(B)
XonStatus   rs.b     1           ;Laatst gestuurde controleteken Xon of Xoff(B)
DummyByte1  rs.b     1           ;Dummy
DummyByte2  rs.b     1           ;Dummy
LEDfunction rs.b     1           ;0=knipper/1=custom (B)
Schecksum   rs.b     1           ;S-record checksum (B)
EchoInput   rs.b     1           ;Wel(1)/niet(0) echo input

SLinenr      rs.w     1          ;S-record regelnummer (W)
SRecOffset   rs.l     1          ;Offset (laden op 0x0000 wordt 0x10400)
SKBsize      rs.w     1          ;Download size in kB (W)
SRecPrgStart rs.l     1          ;Program start address (L)

LeapCounter  rs.l     1          ;Leapcounter for lines/byte etc (L)
FuzixSysCall rs.l     1          ;System call adres fuzix trap-12
FuzixIRQ     rs.l     1          ;IRQ adres fuzix (tbv 100Hz tick)

XModemTeller rs.w    1           ;Telt aantal bytes in 1 blok (128 dus)
XModemSLine  rs.l    1           ;Pointer rnaar S-record regel (l)
XModemCRC    rs.w    1           ;XModem CRC-16 (w)
XModemBloknr rs.w    1           ;XModem blokteller (w)
XModemProto  rs.b    1           ;XModem protocol backup
InUARTbuffer rs.b    1           ;Aantal bytes in UART 16550 buffer 
Status1      rs.b    1           ;Statusbits-1
Status2      rs.b    1           ;Statusbits-2
Wortel       rs.w    1           ;Ruimte voor wortel; priemgetallen(W)
BenchmarkPtr rs.l    1           ;Pointer to benchmark table
HexdumpAddr  rs.l    1           ;Address for hexdump

            even

;LET OP: LineBuf en TemReg overlappen elkaar; bijt elkaar niet
SLineBuff:  rs.b     80          ;LET OP: 80 bytes regelbuffer!!
TempReg:    rs.l     15          ;Ruimte voor alle registers(15xL=60)
TempRegUsp: rs.l     1           ;Ruimte voor USP(L)

;RAM tot 0x010400 SUPER STACK 


; --- INCLUDE ---------------------------------------------------------------
; Include systeem specifieke zaken
           if UART16550
              include "bios_uart16550.inc"
           endif 

           
            org          ROM_offset+$0400  ;BIOS start in EEPROM (after vector table)
; --- INIT -------------------------------------------------------------------
; 23-08-1994 Test supervisor status; bij EPROM start is dat zo, maar bij start
; vanuit RAM niet. LET OP move SR,Dx is privalige in 68010 en hoger, niet 
; gebruiken. Al deze routines gebruikt (nog) geen RAM-geheugen!
MenuB:      clr.l        Ident            ;Initiate cold start (menu-B)
            trap         #1               ;Become SU 
            
Start:      if BIOSinRAM                  ;Startpoint for the BIOS!
            trap         #1               ;Alleen bij start RAM; wordt SU via trap #1
            endif
            move         #$2700,SR        ;Statusregister SU, geen interrupts
            

;Initieer ACIA 6850 (MC68000) op 38.4 kBaud
            if System68K_Classic
            lea          AciaCmd,A0       ;Initieer de 6850 UART
            move.b       #$43,(A0)        ;Reset acia
            move.b       #$55,(A0)        ;Klok delen door 16, 8 bits, geen pariteit
            endif                         ;/geen interrupt, 1 stop, RTS=hoog
 

;Initieer UART 16550 (MC68EC020/Nano68K) op 38.400/115.200 kBaud (SU)
            if UART16550
            lea.l       UART16550,A0
            move.b      #%00001101,FCR(A0)      ; enable FIFO
            move.b      #%10000011,LCR(A0)      ; 8 data bits, no parity, 1 stop bit, DLAB=1
            move.b      #BAUD_DIV_L,DLL(A0)     ; set divisor latch low byte
            move.b      #BAUD_DIV_U,DLM(A0)     ; set divisor latch high byte
            bclr.b      #7,LCR(A0)              ; disable divisor latch
            clr.b       SCR(A0)                 ; clear the scratch register
            move.b      #(1<<MCR_COPI),MCR(a0)  ; SPI COPI idles low

            bset.b      #MCR_LED,MCR(a0)        ;Led aan
            clr.b       InUARTbuffer            ;0 tekens in huidige buffer
            endif
                        
            lea          InitTxt(PC),A1         ;A1=Welcome text
            lea          CopyVector(PC),A2      ;A2=Return address
            bra          StringOut              ;Write text to UART, without RAM


; --- Copy ROM-vectors to RAM ------------------------------------------------
; 2024-09-27 Copy vectors as soon as possible to RAM; made table relocatable
CopyVector: if RAM_offset=0                     ;Copy only if RAM at 0x0000
            lea            VectorTxt(pc),a1     ;A1="Copy vectors to RAM"
            lea            CopyVector1(PC),A2   ;A2=Return adres
            bra            StringOut            ;Write text to UART

CopyVector1:lea            RomVector_00(pc),a0 ;Copy from EEPROM to RAM, 1kB
            movea.l        #$0000,a1           ;RAM/destination at 0x0000
            move.w         #255,d0
            
            lea            Start(PC),a2        ;Reloc vector table
            lea            Start,a3
            suba.l         a3,a2               ;A2=Relocation value
            lea            VectorOkTxt(pc),a4  ;A4=OK-text

CopyVector2:move.l         (a0)+,d1             ;D1=Source EEPROM(a0)
            add.l          a2,d1                ;Calculate new relative address
            move.l         d1,(a1)              ;Store in RAM
            cmp.l          (a1)+,d1             ;Check RAM and increment pointer(A1)
            beq.s          CopyVector3          ;Check if equal
            lea            VectorFailedTxt(pc),a4;A4=Error-text
CopyVector3:dbf            d0,CopyVector2       ;256x longword            

            move.l         a4,a1                ;A1=Vector result text
            lea            Start2(PC),A2        ;A2=Return adres
            bra            StringOut            ;Write text to UART
            endif                               ;RAM_offset=0

            
; --- Cold/warm check --------------------------------------------------------
; Check for cold/warm system start
Start2:     cmpi.l       #MagicNumber,Ident     ;Check cold/warm start
            beq          WarmReset                        


; --- ROM check --------------------------------------------------------------
; 23-08-1994 Bepaal type/grootte (E)EPROM (27C128 2x16kB/2x8KB/2x64kB EEPROM) door
; geheugenplaats $0000 te vergelijken met ghost-geheugenplaatsen
ROMcheck:   lea          ROMcheckTxt(PC),A1  ;Print "-ROM:" zonder RAM!
            lea          ROMcheck0(PC),A2    ;Return adres naar A2
            bra          StringOut           ;Zet starttekst op scherm
            
ROMcheck0:  lea         ROM_offset,A0        ;Start Eprom
            move.l      (A0),D0              ;Haal eerste longwordt uit ROM (=startadres)
            move.l      #1024,D2             ;Test 1,2,4,8..64kB
ROMcheck1:  cmp.l       0(A0,D2.l),D0        ;Zelfde waarde?
            beq.s       ROMcheck2
            lsl.l       #1,D2                ;Test volgende 1>2, 2>4, 4>8 enz
            cmpi.l      #$20000,D2           ;t/m max 64kB
            bne.s       ROMcheck1

ROMcheck2:  lea         0(A0,D2.l),A3        ;Bewaar hoogste adres in A3
            moveq       #10,D0            
            lsr.l       D0,D2                ;Bereken kB's (delen door 1024)
            lea         ROMcheck3(PC),A2     ;Returnadres
            bra         DecimalOut           ;Print aantal kB's       
            
ROMcheck3:  lea         ROMfreeTxt(PC),A1    ;print "...kB,..."
            lea         ROMcheck3a(PC),A2    ;Returnadres
            bra         StringOut
 
;Bepaal aantal vrije bytes (=$FFFF) in ROM
ROMcheck3a: moveq       #-1,D2            ;Aantal bytes=-1 
ROMcheck4:  addq.l      #1,D2             ;Tel aantal lege bytes
            cmpi.l      #$FFFFFFFF,-(A3)  ;Leeg (=FFFF)?
            beq.s       ROMcheck4         ;Totdat gevulde byte gevonden zijn
            
            lsl.l       #2,D2             ;Bereken bytes (x4)
            lea         ROMcheck5(PC),A2  ;Returnadres
            bra         DecimalOut        ;Print aantal bytes free
            
ROMcheck5:  lea         ROMbytesTxt(PC),A1;print "bytes" -tekst
            lea         RAMcheck(PC),A2   ;Return adres
            bra         StringOut


; --- RAM check --------------------------------------------------------------
; 15-08-2021 Controleer de eerste vier bytes van de EVEN en ONEVEN bank
RAMcheck:   lea         RAMevenTxt(PC),A1    ;print "RAM even bank:"
            lea         RAMcheck1(PC),A2     ;Return adres
            clr.l       d7                   ;D7=0: RAM OK 
            bra         StringOut
            
RAMcheck1:  lea         RAM_offset,A0        ;Schrijf even bank AA5500FF
            lea         RAMevenOkTxt(PC),A1  ;Test EVEN bank
            lea         RAMcheck3(PC),A2     ;Return adres            
            move.l      #MagicNumber,D0
            movep.l     D0,0(A0)             ;Schrijf op 4 even bytes
            movep.l     0(A0),D0
            cmpi.l      #MagicNumber,D0
            beq.s       RAMcheck2            ;EVEN RAM OK!
            moveq       #-1,d7               ;RAM fout indicatie in D7
            lea         RAMevenErrorTxt(PC),A1 ;Print "FAILED"
RAMcheck2:  bra         StringOut           


;---Test ONEVEN bank            
RAMcheck3:  lea         RAModdOkTxt(PC),A1   ;Print "odd bank"
            lea         RAMcheck6(PC),A2     ;Return adres
            lea         RAM_offset,A0        ;Schrijf oneven bank AA5500FF
            move.l      #MagicNumber,D0
            movep.l     D0,1(A0)             ;Schijf op 4 oneven bytes
            movep.l     1(A0),D0
            cmp.l       #MagicNumber,D0
            beq.s       RAMcheck5
            lea         RAModdErrorTxt(PC),A1;Print "FAILED"
            moveq       #-1,d7               ;RAM fout indicatie in D7
RAMcheck5:  bra         StringOut    

RAMcheck6:  tst.b       d7                   ;Stop indien RAM niet ok
            bne         Emergency            ;RAM fout; start Emergency operation


; --- MEM check -------------------------------------------------------------
; 22-08-1994 Test max. 128/1024 kB RAM zonder de oorspronkelijk inhoud hiervan te 
; wijzigen. Teller per  16 kB, toon aan einde aantal gevonden kB's.
MEMcheck:   lea          RAM_offset,SP    ;BIOS in ROM; check RAM from beginning
            if BIOSinRAM
            lea          ProgEnd(PC),SP   ;BIOS in RAM: check from the end of BIOS (in RAM)
            endif;
 
MEMcheck0:  clr.l        D2               ;Reset kB-teller.
            lea          MEMtxt(pc),A1    ;Geef melding 'RAM: 0 kB'
            lea          MEMcheck1(PC),A2 ;Returnadres van StringOut
            bra          StringOut        ;zonder gebruik RAM

MEMcheck1:  move.w       #255,D1          ;Test blokjes van 1 kB (=256*longword)
MEMcheck2:  cmpa.l       #RAM_end,SP      ;Einde RAM bereikt?
            bge.s        MEMcheck4        ;Ga niet boven 0x3.0000
            not.l        (sp)             ;Inverteer waarde in RAM
            move.l       (sp),d0
            not.l        d0
            move.l       d0,(SP)          ;Zet de oorspronkelijke waarde terug
            cmp.l        (sp)+,d0
            dbne         d1,MEMcheck2
            bne.s        MEMcheck5        ;OEI! geen werkend ram: klaar.

            if UART16550
            bchg.b       #MCR_LED,UART16550+MCR;Knipper LED
            else
            bclr.b       #PB2_LED,ViaPB   ;Zet de LED aan (=0) 
            endif;

            addq.l       #1,d2            ;Weer 1kB getest
            move.l       d2,d0
            and.w        #$000F,d0        ;Test op volgende 16kB?
            bne.s        MEMcheck1        ;Geen 16 kB? Dan RAM verder testen

            lea          MEMcheck3(PC),A2
            bra          DecimalOut       ;Druk D2 decimaal af

MEMcheck3:  lea          MEMbsTxt(PC),A1  ;CR; cursor naar beginpositie
            lea          MEMcheck1(PC),A2 ;Return adres
            bra          StringOut

MEMcheck4:  lea          MEMcheck5(PC),A2 ;Klaar met RAM-test; druk nog een 
            bra          DecimalOut       ;keer aantal kB's (D2) af.

MEMcheck5:  move.l       sp,MaxMem        ;Zet einde ram in geheugen.
            tst.l        d2               ;Stop BIOS if less than 1 kB RAM
            beq          Emergency


; --- VIA/RTC check ----------------------------------------------------------
; 20-08-2021 Test RTC 1 seconde tick op CA1 (nodig voor calibratie)
            if System68K_Classic          ;Alleen bij MC68000-SBC
RTCcheck:   lea          RTCtxt(pc),A1    ;Geef melding 'RTC: 1 sec tick'
            lea          RTCcheck1(PC),A2 ;Returnadres van StringOut
            bra          StringOut        ;zonder gebruik RAM

RTCcheck1:  clr.b	       ViaDDRA          ;VIA poort-PA wordt input
            clr.b        ViaDDRB          ;VIA poort-PB wordt input
            lea          ViaIFR,A0        ;IFR-status register (interrupts) in A0
            bset         #IFR_CA1,(A0)    ;Clear dit bitje (door 1 te schrijven!)

            lea          RTCerrorTxt(pc),A1 ;Laadt alvast melding 'ERROR'
            move.l       #300000,d0       ;max 300.000 keer testen (circa 1 sec)
RTCcheck2:  subq.l       #1,d0
            beq.s        RTCcheck3        ;GEEN tick gevonden!
            btst         #IFR_CA1,(A0)    ;Controleer IFR bit 1
            beq.s        RTCcheck2  
       
            lea          RTCokTxt(pc),A1 ;Geef melding 'OK'
RTCcheck3:  lea          TimerCheck(PC),A2 ;Returnadres van StringOut
            bra          StringOut        ;zonder gebruik RAM;Tick gevonden            
            endif;
            

; --- VIA/TIMER-1 check ------------------------------------------------------
; 20-08-2021 Test VIA timer-1; calibreer CPU-snelheid. Gebruik memory/interrupts
TimerCheck: move.l       #DefaultClck,CPUclock  ;Standaard 8MHz klok
            if System68K_Classic
            lea          TimerTxt(pc),A1  ;Geef melding 'VIA UM6522: Timer 1...'
            lea          TimerCheck1(PC),A2 ;Returnadres van StringOut
            bra          StringOut        ;zonder gebruik RAM          
            
;Bereken CPU-snelheid via timer-1 (=CPU-klok/10) en 1Hz RTC.
;XXXX LET OP: HIER WORDT MEM GEBRUIKT!
TimerCheck1:bsr          InitTraps        ;Initieer Trap-12 en IRQ/100Hz
            bsr          InitVia          ;Default initiatie VIA

            move.w       #500,d0          ;Delen door 500 (250 mist interrupts)
            move.b       d0,ViaT1LL       ;Programmeer lage deel teller-1
            lsr          #8,d0
            move.b       d0,ViaT1CH       ;Programeer hoge deel teller-1
            move.b       d0,ViaT1LH       ;Is dit nodig???????

            move         #$2000,sr        ;Enable interrupts voor de eerste keer!!!

            bchg         #0,ViaPCR        ;Wissel edge CA1; scheelt .5s wachten)
            clr.l        TimeSec          ;Maak seconde teller nul
            move.l       #300000,d0       ;Wacht op eerste kloktick, max 300.000x 
TimerCheck2:subq.l       #1,d0
            beq.s        TimerCheck5      ;TIMEOUT/ERROR
            tst.l        TimeSec          ;Controleer secondeteller
            beq.s        TimerCheck2  

            clr.l        TimerTicks       ;Zet teller op nul; tel nu halve(!) seconde
            clr.l        TimeSec          ;Zet secondeteller weer op nul
            bchg         #0,ViaPCR        ;Wissel active edge CA1; tel .5 seconde
TimerCheck3:tst.l        TimeSec          ;Wacht op klok-seconde-tick van RTC
            beq.s        TimerCheck3  

            move.l       TimerTicks,d2    ;Reken om naar CPU HZ (*10.000)
            mulu.w       #10000,d2 
            move.l       d2,CPUclock      ;Bewaar deze waarde in CPUclock (Hz)

            divu.w       #1000,d2         ;Print CPU in KHZ; DecimalOut max 99.9999
            lea          TimerCheck4(PC),A2
            bra          DecimalOut       ;Druk D2 decimaal af (MHz)

TimerCheck4:lea          TimerCPUtxt(PC),A1;Print "KHz (OK)"
            lea          InitInt(PC),A2   ;Return adres: init UART
            bra          StringOut

TimerCheck5:lea          TimerErrorTxt(PC),A1;Print "ERROR"
            lea          InitInt(PC),A2   ;Return adres: init UART
            bra          StringOut
            endif;
     
     
; === Calculate CPU-clock ====================================================
; 11-2023 Calculate CPU-clock by writing 16 chars to UART 16550 
            if SystemEC020SBC             ;Enable 020-instruction cache/MC68EC020
            moveq.l      #9,d0            ;2024-02-17 Enable+clear cache on 68020   
            movec.l      d0,CACR
            endif;

            if UART16550
            lea          CPUclock1Txt(PC),A1;Print "CPU: clock"
            lea          CPUclock0(PC),A2  ;Return address
            bra          StringOut

CPUclock0:  lea          UART16550,a0     ;A0:=Base address UART 16550
CPUclock1:  btst         #6,LSR(a0)       ;Check for empty FIFO+shift register
            beq.s        CPUclock1
            
            moveq.l      #7,d0            ;Fill FIFO with 16 chars
CPUclock2:  move.b       #' ',THR(a0)     ;Send space
            move.b       #bs,THR(a0)      ;Send backspace
            dbf          d0,CPUclock2

            clr.l        d0               ;Counter for MHZ *2
CPUclock3:  move.w       #RelCPUspeed*BAUD_DIV_L-3,d1 ;#Loops depends on baudrate and CPU!
CPUclock4:  dbf          d1,CPUclock4     ;Loop count (in cache)
            addq.l       #1,d0            ;Increment counter
            btst         #6,LSR(a0)       ;Check for empty FIFO+shift register
            beq.s        CPUclock3
            
;XXXXXX ONDERSTAANDE ALLEEN VOOR TEST: UITSCHAKELEN            
;            move.l       d0,d2            ;Print nr loops
;            lea CPUstop(pc),a2
;            bra          DecimalOut
;CPUstop:    bra CPUstop ;XXXXXXXXXXXXXXXXXX            
            
            move.l       d0,d2            ;Print D2 as MHz
            mulu.w       #62500,d0        ;Calculate Hz (62500*16=1.000.000)
            lsl.l        #4,d0            ;d0:=d0 * 16
            move.l       d0,CPUclock      ;Store results

            lea          CPUclock5(PC),A2
            bra          DecimalOut       ;Druk D2 decimaal af (MHz)

CPUclock5:  lea          CPUclock2Txt(PC),A1;Print "MHz"
            lea          CPUclock6(PC),A2  ;Return address
            bra          StringOut

CPUclock6:            
            endif;        
                       
            
;=== 100 Hz oscilator check, only Thunderbird and ByteBlazer (020) ===========
            if (SystemEC020SBC|System68K_Thunderbird)
Oscillator: lea          OscillatorTxt(PC),A1;"UART: test 100Hz oscillator"
            lea          Osc1(PC),A2      ;Return adres: init UART
            bra          StringOut            

Osc1:       lea          OscillatorErrorTxt(PC),a1;Prepair error text
            lea          UART16550,a0     ;A0:=Base address UART
            moveq.l      #-1,d0           ;Try 65.535x
Osc2:       btst         #MSR_RI,MSR(a0)  ;check ringindicator(=0)
            dbeq         d0,Osc2
            bne.s        OscError         ;No level switch ringindicator/100Hz
            
            moveq.l      #-1,d0           ;Try 65.535x
Osc3:       btst         #MSR_RI,MSR(a0)  ;check ringindicator(=1)
            dbne         d0,Osc3
            beq.s        OscError         ;No level switch ringindicator/100Hz
                        
            lea          OscillatorOkTxt(PC),a1;Write OK text
OscError:   lea          InitInt(PC),A2   ;Return adres: init UART
            bra          StringOut            
            endif;
            
            
; --- INIT Interrupts --------------------------------------------------------
; 1991: Zet UART & VIA (timer) interrupts aan, daarna het MENU. Is tevens warme start!
InitInt:    clr.l        TimerTicks       ;Zet de tijd na een koude start op nul
            lea          UARTtxt(PC),A1   ;Print "-SYS: Enable RAM"
            lea          InitInt0(PC),A2  ;Return adres: init UART1
            bra          StringOut

; Now we can use RAM. Enable 68020 instruction cache
InitInt0:
            if SystemEC020SBC
            lea          UARTtxt1(PC),a1  ;Print "(OK), enable 020-instruction cache"
            bsr          StrOutT15
            moveq.l      #9,d0
            movec.l      d0,cacr          ;Clear & enable 68020 instruction cache
            endif
                 
            lea          UARTtxt2(PC),a1  ;Print "(OK), enable ISR"
            bsr          StrOutT15
       
  
; SU-modes. Zet als eerste eventuele resten van Fuzix uit!
WarmReset:                                ;Beginpunt warme reset
InitInt1:   bsr          InitTraps        ;Initiëer TRAP-12 SysCall en IRQ/100Hz
            bsr          InitUART         ;(Her)programmeer UART met interrupts
            bsr          InitVia          ;Programmeer timer-1 via berekende CPUclock
            bsr          SRecInit         ;Init program start address
            bsr          InitCPUcache     ;Switch on instruction cache

            movea.l      MaxMem,a0        ;Zet USP op hoogste memory adres (0x3.0000)
            move.l       a0,usp           ;Gebruikers SP initieeren
            lea          SStack,SP        ;Geef super SP een nieuwe waarde 0x10400
            st           EchoInput        ;Bij regelinvoer input echo
          
            move         #$0000,SR        ;Alle INT aan & MENU uitvoeren in USERMODE!

            cmpi.l       #MagicNumber,Ident;Test op warme/koude start, dan naar Menu
            beq          Menu
            move.l       #MagicNumber,Ident

            lea          UARTokTxt(PC),a1 ;Print "OK" van de koude start/UART test
            bsr          StrOutT15
            bra          Menu
 
 
; --- INIT VIA ---------------------------------------------------------------
; 1993 Programmeer de poorten en VIA-timers;
; -Poorten: PA+PB input (LED aan)
; -Timer-1: 100 Hz tick (vast, niet meer programmeerbaar)
; -Timer-2: uit. (Deels te programmeren via bios trap #3)
InitVia:	   if System68K_Classic
            clr.b	       ViaDDRA          ;Zet de poorten van de VIA op input
            clr.b	       ViaDDRB
            clr.b        LEDfunction      ;LED knipperen

            bset.b       #PB2_LED,ViaDDRB ;PB2/gele LED=output
            bclr.b       #PB2_LED,ViaPB   ;Zet de LED aan (=0) 

            move.b	    #$40,ViaACR      ;Timer-1 freerun, timer-2 single shot

            move.l       CPUclock,d0      ;Haal CPU-klok
            divu.w       #10*Timer1Freq,d0;Bereken deler tov CPU-klok/10
            cmp.w        #1000,d0         ;Kleiner dan 1000 is te snel/fout
            bge.s        InitVia1
            move.w       #8000,d0         ;Neem dan 8000 (8MHz klok/100Hz tick)
InitVia1:   move.b       d0,ViaT1LL       ;programmeer de teller timer-1.
            lsr          #8,d0
            move.b       d0,ViaT1CH
            move.b       d0,ViaT1LH       ;Is dit nodig???
;Om de interrupts aan te zetten moet je bit 7 van IER op 1, en vervolgens de bit
;die je aan wil zetten. Omdat IER bit 7 bij lezen altijd 1 is; kun je bset gebruiken!
            bset         #IER_tmr1,ViaIER ;Enable interrupts timer-1 (bit 6)
            bset         #IER_CA1,ViaIER  ;Enable interrupts CA1 (bit 1)
            endif;
            rts

            
; --- Init UART --------------------------------------------------------------
; 1991: Initieer de ACIA/UART voor gebruik met interrupts. Aanroep moet in SU zijn!
InitUART:   move.w       sr,-(sp)         ;Bewaar huidige statusregister
            move.w       #$2700,sr        ;Even alle interrupts uit

            clr.b        Status1          ;Alle statusbitten op nul
            clr.b        Status2          ;Alle statusbitten op nul

            clr.w        RecLeesPtr       ;Zet lees- en schrijfpointers van de
            clr.w        RecAantal        ;yclische buffers op nul
            st           ProtocolOn	      ;Xon/Xoff-protocol
;            sf           SysReset         ;Niet bezig met esc-@ (reset)
;            sf           RecEsc
            st           XonStatus        ;Xon aanzetten 
            
            if System68K_Classic
            move.b       #AciaTrOff,AciaCmd ;Programmeer ACIA 6850 met interrupts
            move.b       #AciaTrOff,XonXoff
            endif;

            if UART16550
            lea          UART16550,a0     ;A0:=base address UART16550
;           move.b       #$81,FCR(a0)     ;FIFO on, int at  8 char in FIFO          
            move.b       #$C1,FCR(a0)     ;FIFO on, int at 14 char in FIFO          
            move.b       #$09,IER(a0)     ;Interupts RX on (bit 0) & modem on (bit 3)
;           move.b       #$08,IER(a0)     ;Interupts on for modem (bit 3)
            endif;
            
            move.w       (sp)+,sr         ;Restore statusregister
            rts
            
 
;=== InitTraps ===============================================================
; 05-02-2022 Initialize traps by RAM (Trap-12 + 100Hz IRQ)
InitTraps:  movem.l        a0-a3/d0,-(sp)   ;Push A0/A1/D0
            lea            NullProc(pc),a0  ;NullProc=RTE
            move.l         a0,FuzixSysCall  ;Stel NullProc in voor SysCall/Trap-12
            move.l         a0,FuzixIRQ      ;Stel NullProc in voor IRQ/100Hz

            lea            RomVector_00(pc),a0 ;Copy from ROM to RAM, 1kB
            movea.l        #$0000,a1         ;RAM/destination at 0x0000
            move.w         #255,d0
            lea            Start(PC),a2      ;Reloc vector table
            lea            Start,a3
            suba.l         a3,a2
InitTraps1: move.l         (a0)+,d1          ;Copy from ROM(a0) to RAM(a1)
            add.l          a2,d1             ;Calculate new relative address
            move.l         d1,(a1)+          ;Store in RAM
            dbf            d0,InitTraps1     ;256x longword            

            movem.l        (sp)+,a0-a3/d0    ;Pop A0/A1/D0
            rts

             
;--- InitCPUcache ---------------------------------------------------------------
;10-12-2023 Init instruction cache 68020 (on 68000 do nothing) & print message
InitCPUcache:  
            if SystemEC020SBC
            moveq.l      #T15CacheCtrl,d0 ;Enable(1) instruction cache
            moveq.l      #1,d1            
            trap         #15
            endif;
            rts

             
;--- STRINGOUT ---------------------------------------------------------------
; Functie:  String (A1) naar UART 6551; #0 = einde string zonder RAM gebruik.
; In:       A1=string, A2=returnadres
; Uit:      Tekst(A1) naar UART
; Gebruikt: A0, A1
; Datum:    07-1991 
; LET OP:   Maak A0 reloceerbaar via "LEA txt(PC),A0"
            if System68K_Classic
StringOut:  lea          AciaCmd,A0    ;A0 wijst naar UART-command register
StringOut1: btst.b       #1,(A0)       ;Wacht totdat de UART klaar is
            beq.s        StringOut1

            move.b       (A1)+,2(A0)   ;Print een teken naar AciaData
            tst.b        (A1)          ;Test op einde string (#0)
            bne.s        StringOut1
            addq.l       #1,A1         ;A1 verwijst naar volgende string
            jmp          (A2)          ;Spring terug zonder RAM gebruik
            endif
            
            if UART16550
StringOut:  lea          UART16550,A0  ;A0 wijst naar UART-base address
StringOut1: btst.b       #5,LSR(A0)    ;Wacht totdat de UART klaar is
            beq.s        StringOut1

            move.b       (A1)+,THR(A0) ;Print een teken naar AciaData
            tst.b        (A1)          ;Test op einde string (#0)
            bne.s        StringOut1
            addq.l       #1,A1         ;A1 verwijst naar volgende string
            jmp          (A2)          ;Spring terug zonder RAM gebruik
            endif
            
            
;--- DECIMALOUT --------------------------------------------------------------
; Functie:  Druk D2.W af zonder RAM gebruik
; In:       D2.w=getal, A2=returnadres
; Uit:      D2.w decimaal naar UART 6551
; Gebruikt: A0, D5, D6 & D7
; Datum:    22-08-1994 
DecimalOut: if System68K_Classic
            lea          AciaCmd,a0      ;MC68000: adres UART-command register naar A0
            elseif
            lea          UART16550,a0    ;MC68EC020: Base adres UART
            endif
            
            clr.l        d6
            move.w       d2,d6           ;Reken met d6
            move.l       #10000,d5       ;D5 deler
            moveq        #' ',d7         ;D7=te printen char/cijfer
            
Decim0:     divu.w       d5,d6           ;Duizendtal in D6
            beq.s        Decim1
            moveq        #'0',d7
Decim1:     add.b        d7,d6           ;Maak ascii
            if System68K_Classic
Decim2:     btst.b       #1,(A0)         ;Wacht tot acia vrij is
            beq.s        Decim2
            move.b       d6,2(A0)        ;Print duizendtal
            endif
            
            if UART16550
Decim2:     btst.b       #5,LSR(A0)      ;Wacht tot UART vrij is
            beq.s        Decim2
            move.b       d6,THR(A0)      ;Print duizendtal
            endif
            
            
            clr.w        d6              ;Rest staat in D6-hoog!
            swap         d6
            divu         #10,d5          ;deler 10.000>1000>100>10>1>0
            beq.s        Decim4          ;Deler=0? Klaar
            cmpi         #1,d5           ;Laatste cijfer, dan D7 wordt "0"
            bne.s        Decim0
            moveq        #'0',d7
Decim3:     bra.s        Decim0
Decim4:     jmp          (A2)            ;Terug naar aanroeper!



; === MENU ===================================================================
; 1991: BIOS menu with benchmarks, tests and loading programs
; 2024-10-02 No spaces allowed around minus; calculation in vasm will go wrong!
MenuJmpTbl: dc.w        MenuA-MenuJmpTbl  ;A: Access violation (trap)
            dc.w        MenuB-MenuJmpTbl  ;B: Boot system
            dc.w        MenuC-MenuJmpTbl  ;C: CHK-error (trap)
            dc.w        MenuD-MenuJmpTbl  ;D: Division by zero (trap)
            dc.w        MenuE-MenuJmpTbl  ;E: Emergency operation
            dc.w        MenuF-MenuJmpTbl  ;F: Fibonacci benchmarks
            dc.w        Menu-MenuJmpTbl   ;G: Voided
            dc.w        MenuH-MenuJmpTbl  ;H: Hexdump
            dc.w        MenuI-MenuJmpTbl  ;I: Illegal (trap)
            dc.w        Menu-MenuJmpTbl   ;J: Voided
            dc.w        Menu-MenuJmpTbl   ;K: Voided
            dc.w        MenuL-MenuJmpTbl  ;L: Load program (old)
            dc.w        MenuM-MenuJmpTbl  ;M: Memory map 
            dc.w        Menu-MenuJmpTbl   ;N: Voided
            dc.w        Menu-MenuJmpTbl   ;O: Voided
            dc.w        MenuP-MenuJmpTbl  ;P: Primes
            dc.w        Menu-MenuJmpTbl   ;Q: Voided
            dc.w        MenuR-MenuJmpTbl  ;R: Run program
            dc.w        MenuS-MenuJmpTbl  ;S: S-loader
            dc.w        MenuT-MenuJmpTbl  ;T: Set time (RTC)
            dc.w        Menu-MenuJmpTbl   ;U: Voided
            dc.w        Menu-MenuJmpTbl   ;V: Voided
            dc.w        Menu-MenuJmpTbl   ;W: Voided
            dc.w        MenuX-MenuJmpTbl  ;X: X-modem
            dc.w        Menu-MenuJmpTbl   ;Y: Voided
            dc.w        Menu-MenuJmpTbl   ;Z: Voided


; --- MENU -------------------------------------------------------------------
MenuHelp:   lea         MenuTxt(pc),a1    ;Zet menu(help)tekst op het scherm
            bsr         StrOutT15
Menu:       lea         CommandTxt(pc),a1 ;Zet "Command" op het scherm
            bsr         StrOutT15
            
Menu1:      bsr         CharInT15         ;D1:=command
            andi.w      #$00FF,d1         ;Convert byte to word
            cmpi.w      #'a',d1           ;Convert to uppercase
            blo.s       Menu2
            cmpi.w      #'z',d1
            bhi.s       Menu2
            subi.w      #'a'-'A',d1
            
Menu2:      subi.w      #'A',d1           ;Check range A-Z
            blo.s       Menu              ;Wrong character (<A); show menu text
            cmpi.w      #25,d1         
            bhi.s       Menu              ;Wrong character (>Z): show menu text

            add.w       d1,d1             ;D1:=D1*2 
            lea         MenuJmpTbl(PC),a1 ;A1:=start jump table
            move.w      (a1,d1.w),d1      ;Relocate address from table
            adda.w      d1,a1             ;A1:=points to routine from jump table
            jsr         (a1)              ;Jump to subroutine
            bra.s       Menu              ;Back to menu again

            
; --- Menu-A: Access voilation (trap) ----------------------------------------
MenuA:      if SystemEC020SBC             ;Alternate cache status on 68020
            moveq        #T15CacheCtrl,d0 ;Query cache status
            moveq        #-1,d1
            trap         #15
            
            moveq        #T15CacheCtrl,d0 ;Set cache status
            eor.w        #1,d1            ;Toggle cache bit
            trap         #15              ;Set cache status

            lea          CacheEnableTxt(PC),a1
            btst         #0,d1
            bne.s        MenuA1
            lea          CacheDisableTxt(PC),a1
MenuA1:     moveq        #T15StrOut,d0
            trap         #15              ;Show cache status
            rts

            else                          ;Initiate address error
            move.w       $87654321,D1     ;This works only on 68000/68010
            rts
            endif;


; --- Menu-C: rangecheck (trap) ----------------------------------------------
MenuC:      moveq.l      #123,d0
            chk          #100,d0          ;Initiate CHK (trap)

            
; --- Menu-D: Division by zero (trap) ----------------------------------------
MenuD:      clr.l        d0
            divu.w       d0,d1            ;Initiate Division-by-zero (trap)


; --- Menu-I: Illegal (trap) -------------------------------------------------
MenuI:      illegal                       ;Initiate illegal (trap)


; --- Menu-M: Memory map -----------------------------------------------------
MenuM:      lea          MemoryMapTxt(PC),a1
            moveq        #T15StrOut,d0
            trap         #15              ;Show the memory map
            rts

; --- Menu-T: Set time (RTC) -------------------------------------------------
MenuT:      bsr          ShowTime         ;Show RTC-time
            bsr          SetTime          ;Set RTC-time
            bra          ShowTime         ;Show RTC-time again/return to menu



; --- Alternate instruction cache --------------------------------------------
            if SystemEC020SBC             ;Alternate cache status (68020 only)
AlternateCache:
            moveq        #T15CacheCtrl,d0 ;Query cache status
            moveq        #-1,d1
            trap         #15
            
            moveq        #T15CacheCtrl,d0 ;Set cache status
            eor.w        #1,d1            ;Toggle cache bit
            trap         #15

            lea          CacheEnableTxt(PC),a1
            btst         #0,d1
            bne.s        AlternateCache1
            lea          CacheDisableTxt(PC),a1
AlternateCache1:
            moveq        #T15StrOut,d0
            trap         #15
            
            rts
            endif;
            
            
; --- DOWNLOAD ----------------------------------------------------------------
; 1991: Laad programma <high><low><n*data><xor>
; D2=CRC, D3=lengte programma, A0=Download adres
MenuL:
DownLoad:   bsr          SRecInit       ;Clear start address
            clr.b        ProtocolOn     ;Zonder Xon/Xoff protocol
            lea          LaadTxt(pc),A1
            bsr          StrOutT15      ;Print melding

            bsr          CharInT15      ;Lees <HI>-byte in D1
            lsl          #8,D1          ;D1.w=lengte programma
            move.w       d1,-(sp)       ;Bewaar op stack

            bsr          CharInT15      ;Lees <LO>-byte in D1
            andi.w       #$00FF,d1
            move.w       d1,d3
            add.w        (sp)+,d3       ;D3=lengte programma
            
            clr.l        D2             ;D2 = Checksum.
            lea          Program,A0     ;A0 = Programma
            bra.s        Download1

Download0:  bsr          CharInT15      ;Lees n*data
            move.b       d1,(A0)+       ;Byte in geheugen
            eor.b        d1,d2          ;Bereken checksum
Download1:  dbf          D3,Download0

            bsr          CharInT15      ;Checksum ok?
            st           ProtocolOn     ;Xon/Xoff protocol weer aan.
            cmp.b        d1,d2
            beq          SPrgStart      ;Start programm
            
Download3:  lea          LaadErrTxt(pc),a1 ;Checksum wrong
            bra          StrOutT15      ;Print errormessage; return to menu


; --- ShowTime ---------------------------------------------------------------
; 20-11-2021 Toon de tijd (HH:MM:SS)
ShowTime:   movem.l     d0-d3/a1,-(sp) ;Bewaar registers
            lea         TimeShowTxt(pc),a1  ;Print "Current time"
            moveq       #T15StrOut,d0
            trap        #15
 
            move.l      TimeSec,d2     ;Toon de tijd in uren, minuten & sec
            
            move.l      #60*60*24,d1   ;Toon aantal dagen
            moveq       #' ',d0        ;Print na dagen een spatie
            bsr         ShowTimePart
         
            move.l      #60*60,d1      ;Toon aantal uren
            moveq       #':',d0        ;Print na dagen een spatie
            bsr         ShowTimePart

            move.l      #60,d1         ;Toon aantal minuten
            moveq       #':',d0        ;Print na dagen een spatie
            bsr         ShowTimePart           
 
            moveq       #1,d1          ;Toon aantal secondes
            moveq       #' ',d0        ;Print spatie na #secondes
            bsr         ShowTimePart

            movem.l     (sp)+,d0-d3/a1 ;Red registers
            rts


; --- ShowTime ---------------------------------------------------------------
; Print d2 div d1; rest blijft in d2. D0 bevat extra teken (bv :) Bij gebrek 
; aan DIV.L delen via afrekken.
; Gebruikt d0,d1,d2
ShowTimePart:
            move.l      d0,-(sp)       ;Bewaar extra teken (:)
            moveq       #-1,d3         ;Bereken: d3:=d2 div d1; d2:=d2 mod d1
 
ShowTimeP1: addq.l      #1,d3
            sub.l       d1,d2 
            bpl.s       ShowTimeP1
            add.l       d1,d2
 
            cmpi        #10,d3         ;Print extra "0" indien <=9
            bge         ShowTimeP2

            moveq.l     #T15CharOut,d0 ;Print "0"
            moveq       #'0',d1
            trap        #15

ShowTimeP2: moveq       #T15DecOut,d0  ;Decimal out d1
            move.l      d3,d1
            trap        #15

            move.l      (sp)+,d1       ;Print extra teken (:)
            moveq       #T15CharOut,d0 ;Charout via trap #15
            trap        #15
            rts

            
; --- SetTime ----------------------------------------------------------------
; 21-11-2021 Stel de tijd in van de 1-sec timer (HH:MM)
SetTime:    movem.l      d0/d1/a1,-(sp) ;Bewaar registers d0/d1/a1
            lea          TimeSetTxt(pc),a1  ;Print "Set time"
            moveq        #T15StrOut,d0
            trap         #15

            lea          TempReg,a1    ;Lees null terminated string in A1
            moveq        #T15ReadStr,d0
            trap         #15
            
            lea          TempReg,a1    ;Lees null terminated string in A1
            bsr          DecValue      ;Haal uren uit (A1) en zet in d0.l
            cmpi.l       #23,d0        ;Test op 23, fout indien groter
            bgt          SetTimeFout
            move.l       d0,d1

            bsr          DecValueSkip  ;Haal minuten uit (A1) en zet in d0.l
            cmpi.l       #59,d0        ;Test op 59; fout indien groter      
            bgt          SetTimeFout
            mulu         #60,d1
            add.l        d0,d1
            
            bsr          DecValueSkip  ;Haal secondes uit (A1) en zet in d0.l
            cmpi.l       #59,d0        ;Test op 59; fout indien groter
            bgt          SetTimeFout
            mulu         #60,d1
            add.l        d0,d1
            
            tst.l        d1            ;00:00:00 beschouwen we als fout
            beq.s        SetTimeFout
            move.l       d1,TimeSec
SetTimeFout:            
            movem.l      (sp)+,d0/d1/a1 ;Herstel registers
            rts

            
; --- DecValue ---------------------------------------------------------------
; 04-12-2021 Lees decimaal getal (A1) in D0.l. 
; D1=teken. Hoog (A1) telkens op. 
; Opm: ALLEEN POSITIEVE GETALLEN (tzt UITBREIDEN?)
DecValueSkip:tst.b       (a1)          ;Skip één teken (indien niet einde string)
            beq          DecValue
            addq.l       #1,a1         ;Skip teken onder (A1)

DecValue:   move.l       d1,-(sp)      ;Bewaar d1
            clr.l        d0            ;Maak resultaat D0 alvast nul

DecValue1:  clr.l        d1            ;d1.l wordt teken uit (A1)
            move.b       (a1),d1       ;Test input; einde string?
            beq          DecValue2
            
            subi.b       #'0',d1       ;Kleiner dan 0?
            blt.s        DecValue2
            cmpi.b       #9,d1         ;Groter dan 9?
            bgt.s        DecValue2
            
            lsl.l        #1,d0         ;Reken d0 := d0 * 10 + d1
            move.l       d0,-(sp)
            lsl.l        #2,d0         ;d0:=d0*8
            add.l        (sp)+,d0      ;d0:=d0*8 + d0*2
            add.l        d1,d0         ;Eindresultaat weer in d0.l

            addq         #1,a1         ;Volgende positie
            bra          DecValue1     ;Naar volgende teken

DecValue2:  move.l       (sp)+,d1      ;Herstel d1/A0
            rts
       
            
; === TRAP-0 =================================================================
; 1991: Stop user-program/show message/initialize stackpointers/goto Menu 
Trap0EndProgram:
            lea            SStack,SP      ;Initialize SSP (0x10400)
            movea.l        MaxMem,a0      ;Initialize USP (0x30000)
            move.l         a0,usp

            bsr            InitTraps      ;Init RAM-vectors (0x0000)
            
            lea            EindPrTxt(pc),A0 ;Print "Program terminated"
            bsr            StrOut

            bsr            EmptyBuffer    ;Truncate inputbuffer 
            
            bsr            CharIn         ;<Hit any key> to proceed
            bra            Menu           ;Ga naar BIOS-menu



; === EXCEPTION ==============================================================
; 1991 Exception afhandeling bij een fout; zet de PC en registers op het scherm
IllErr:     pea          InstTxt(pc)    ;Startadres van 'Illegal inst'-tekst
            bra.s        Error
DivErr:     pea          DivTxt(pc)
            bra.s        Error
ChkErr:     pea          ChkTxt(pc)
            bra.s        Error
OvErr:      pea          TrapVTxt(pc)
            bra.s        Error
PrivErr:    pea          PrivTxt(pc)
            bra.s        Error
TraceErr:   pea          TraceTxt(pc)
            bra.s        Error
SpurErr:    pea          SpurTxt(pc)
            bra.s        Error
Reset0:     move         #$2000,SR     ;Laat alle int weer door.
;            tst.b        SysReset
;            bne          Start        ;Geef een reset
            pea          ResetTxt(pc)
            bra.s        Error         ;Zet reset tekst op het scherm

AddrErr:    pea          AdrTxt(pc)    ;Address-error exception
            bra.s        ABError

BusErr:     pea          BusTxt(pc)    ;Buserror exception
; LET OP: bij Adres-erros is stackframe anders.
; PC(l) / SR(w) / Inst(w) / Adres(l) / Functcode(w)
ABError:    movem.l      D0-A7,TempReg ;Zet registers in geheugen
            bsr          InitTraps     ;Zet alle restjes van FUZIX uit
            lea          AbTxt(pc),A0  ;'Access to'
            bsr          StrOut

            move.l       (SP)+,A1      ;Tekstwijzer 'bus/adres' naar A1
            addq         #2,SP         ;Verwijder functioncode(W) van stack
            move.l       (SP)+,D0      ;Pop Acces address
            bsr          HexAddrOut

            bsr          StrOut        ;'caused a'

            move.l       A1,A0         ;'bus/address'
            addq         #2,SP         ;Verwijder instr register
            bra.s        Error1

            
; --- ERROR ------------------------------------------------------------------
; 1991 Toon alle registers na een ILLEGAL/Address enz trap
Error:      movem.l      D0-A7,TempReg    ;Zet alle registers incl SSP in het geheugen
            bsr          InitTraps        ;Zet alle restjes van FUZIX uit
            lea          Err0Txt(pc),A0   ;Zend een Xon en een GotoXY.
            bsr          StrOut

            move.l       (SP)+,A0         ;Haal de tekstverwijzing op.
Error1:     bsr          StrOut           ;Zet exception-tekst op het scherm
            lea          ErrTxt(pc),A0    ;' error at PC '
            bsr          StrOut
 
            move.w       (sp)+,d7         ;Pop statusregister (SR) in D7
            andi.w       #$2000,d7        ;Test op SU (0x2000 = SU)
            bne.s        Error2
            move.l       usp,a1           ;Zet USP ipv SSP in TempReg
            move.l       a1,TempRegUsp
            
Error2:     move.l       (SP)+,D0         ;Pop PC-address (D0) en print
            bsr          HexAddrOut
            
            bsr          StrOut           ;Print ' (='
            andi         #$FFFE,d0        ;Maak adres even
            movea.l      d0,a1
            move.w       (a1),d0          ;Haal instructie op
            bsr          HexWrdOut


            lea          RegTxt(pc),A0    ;Zet alle register op het scherm
            lea          TempReg,A1
            move         #15,D1
RegLoop:    bsr.s        StrOut           ;'RegX = '
            move.l       (A1)+,D0
            bsr.s        HexOut
RegLp2:     dbf          D1,RegLoop
            bsr.s        StrOut

            trap         #0               ;End program/warm reset


; --- HEXOUT -----------------------------------------------------------------
; 1991 Print D0.l hexadecimaal, in woord (4)/Adres (6) of long(8)
; D0=teken uit, D1=#bits te schuiven, D2=rest te printen getal 
HexByteOut: movem.l      D0-D2,-(SP)   ;4 cijfers voor woord
            moveq        #4,D1         ;4x schuiven naar rechts voor meest sign nibble
            bra.s        HexnOut

HexWrdOut:  movem.l      D0-D2,-(SP)   ;4 cijfers voor woord
            moveq        #12,D1        ;12x schuiven naar rechts voor het meest sign nibble
            bra.s        HexnOut

HexAddrOut: movem.l      D0-D2,-(SP)   ;6 cijfers voor adressen
            moveq        #20,D1        ;20x schuiven naar rechts voor het meest sign Nibble
            bra.s        HexnOut

HexOut:     movem.l      D0-D2,-(SP)   ;8 cijfers voor datareg
            moveq        #28,D1

HexnOut:    move.l       D0,D2         ;Bewaar D0 in D2
Hex1:       move.l       d2,d0
            lsr.l        D1,D0
            andi.w       #$000F,D0
            addi.w       #'0',D0
            cmpi.w       #'9',D0       ;Groter dan 9?
            bls.s        Hex2
            add.w        #7,D0         ;Maak dan A t/m F
Hex2:       move.l       d1,-(sp)      ;Bewaar D1 XXXX DIT MOET BETER KUNNEN!!!
            move.l       d0,d1
            bsr          CharOutT15    ;Output D1/nibble
            move.l       (sp)+,d1
            subq         #4,D1
            bpl.s        Hex1          ;Klaar ?
            movem.l      (SP)+,D0-D2
            rts


; --- STROUT -----------------------------------------------------------------
; 1991 Print nul terminated string (A0). A0 verwijst na afloop naar het eerste 
; teken NA de het nulteken, zodat de volgende string geprint kan worden.
; LET OP: Maak A0 reloceerbaar via "LEA txt(PC),A0"
; LET OP: Alleen in supermodus aanroepen
StrOut:     move.l       D0,-(SP)         ;Bewaar D0
StrOut1:    move.b       (A0)+,D0         ;Haal teken
            beq.s        StrOut2          ;#0 Einde string (nieuw)
            bsr.s        CharOut
            bra.s        StrOut1
StrOut2:    move.l       (SP)+,D0         ;Haal D0 weer van stack
            rts


; --- CHAROUT ----------------------------------------------------------------
; 1991 Print teken D0.b; direct als UART/buffer leeg is, zet anders in buffer.
; LET OP: alleen in SUPER-modus aanroepen
            if System68K_Classic
CharOut:    move.l       a0,-(sp)         ;Push a0
            lea          AciaCmd,A0       ;A0 wijst naar UART-command register
CharOut1:   btst.b       #1,(A0)          ;Wacht totdat de UART klaar is
            beq.s        CharOut1
            move.b       d0,2(A0)         ;Print een teken naar AciaData
            move.l       (sp)+,a0         ;Pop a0
            rts
            endif;

; 24-09-2023 UART 16550 heeft 16 bytes buffer; buffer-vol niet te bepalen.
; Houd daarom zelf teller bij
            if UART16550                  ;Voorlopig direct naar UART zonder interrupts
CharOut:    move.l       a0,-(sp)         ;Bewaar A0
            lea          UART16550,A0     ;A0 wijst naar UART-base address
CharOut1:   btst.b       #5,LSR(A0)       ;Controleer of de UART buffer leeg is
            beq.s        CharOut2         ;Spring indien niet leeg
            clr.b        InUARTbuffer     ;UART buffer is leeg; maak teller leeg
CharOut2:   cmpi.b       #16,InUARTbuffer ;Minder dan 16 tekens in buffer?
            bge.s        CharOut1         ;UART buffer vol; wacht op leeg
            move.b       d0,THR(A0)       ;Print D0 naar UART
            addq.b       #1,InUARTbuffer  ;Eén teken meer in de buffer
            
            move.l       (sp)+,a0         ;Redt A0 van stack
            rts
            endif                         ;UART16550


; --- CHARIN(D0) -------------------------------------------------------------
; 1991 CharIn (D0) en CharOut mogen alleen in de supermode worden aangeroepen.
; Gebruik bijvoorkeur Trap #15
CharIn:     movem.l      d1/a0,-(SP)      ;Bewaar registers
            bra.s        CharIn1
            
CharIn0:    STOP         #$2000           ;Wacht op interrupts
CharIn1:    tst.w        RecAantal        ;Wacht tot er een teken in de buffer staat
            beq.s        CharIn0          ;Geen tekens; STOP & wacht op INT

            move.w       #$2700,SR        ;Stop acia-interrupts
            move.w       RecLeesPtr,d1    ;Haal teken uit buffers
            lea          RecBuf,A0     
            move.b       0(A0,D1.w),D0 Haal teken uit buffer.
            addq.w       #1,d1            ;Verhoog leespointer
            andi.w       #RecBufLen-1,d1  ;Buffer is cyclisch dus AND
            move.w       d1,RecLeesPtr

            subq.w       #1,RecAantal     ;Verlaag aantal in buffer

            move.w       #$2000,SR        ;Interrupts weer aan
            movem.l      (SP)+,d1/A0
            rts


; 10-06-2023 Voorlopige direct UART 16550 uitlezen (zonder interrupts)            
;            if UART16550
;CharIn:     btst         #0,UART16550+LSR ;Character received?
;            beq.s        CharIn
;            clr.l        d0
;            move.b       UART16550,d0     ;receive byte and write to D0.b
;            rts
;            endif
            
          
            
; --- ISR:AICAINT ------------------------------------------------------------
; 1991 UART-interrupt routine, gezamelijk UART & VIA (Level-4)
AciaInt:    movem.l      a0/d0/d1,-(SP);Bewaar registers
            lea          AciaCmd,A0
            move.b       (A0),d0       ;Identificeer interrupt
            bpl          ViaInt        ;Geen Aciaint, dan ViaInt

            lsr          #1,d0         ;Teken ontvangen?
            bcs.s        RecData
;            lsr          #1,d0         ;Teken zenden?
;            bcs          TrData
            bra.s        AciaIntRet


RecData:    move.b       2(A0),d0      ;Teken van acia naar d0.
;           cmpi.b       #'@',d0
;           bhi.s        RecData2      ;"Gewoon" teken
;           beq          ResetCmd      ;'@' dus mogelijk resetcommando.

RecData1:   cmpi.b       #Xoff,d0      ;Xoff --> Zet transmit-int uit
            beq.s        TrIntOff
            cmpi.b       #Xon,d0       ;Xon  --> Zet transmit-int aan
            beq.s        TrIntOn

RecData2:   ;move.b       d0,RecEsc     ;Bewaar het laatste teken
            lea          RecBuf,A0     ;Zet data in de buffer
            move         RecLeesPtr,d1
            add          RecAantal,d1  ;Bereken schrijfptr
            andi         #RecBufLen-1,d1
            move.b       d0,0(A0,d1.w) ;Zet teken in buffer
            addq         #1,RecAantal  ;Hoog aantal tekens in buffer met 1 op.
            cmpi         #Bovengrens,RecAantal  ;Wordt het tijd voor een Xoff?
            bne.s        AciaIntRet

            tst.b        XonStatus     ;Zend alleen een Xoff als status Xon is
            beq.s        AciaIntRet
            sf           XonStatus     ;Status:=Xoff
            lea          AciaCmd,A0
ZendXoff:   btst.b       #1,(A0)       ;Zend een Xoff; polling. Op int-basis
            beq.s        ZendXoff      ;geeft dit te veel trammelant en tierelier
            move.b       #Xoff,2(A0)
AciaIntRet: movem.l      (SP)+,a0/d0/d1 ;Klaar
            rte


;XXXXX DIT KAN VERVALLEN; ZENDEN DOEN WE ZONDER INTERRUPT
TrIntOn     move         #AciaTrOff,d1  ;Zet zend-interrupt van acia aan.
            bra.s        TrInt
TrIntOff    move         #AciaTrOff,d1 ;Zet zend-interrupt van acia uit.
TrInt       tst.b        ProtocolOn    ;Werken we wel met het Xon/Xoff protocol?
            beq.s        RecData2      ;Zo niet dan teken alsnog in de buffer

            move.b       d1,XonXoff    ;Bewaar commando in XonXoff
            move.b       d1,(A0)       ;Programmeer acia.
            bra          AciaIntRet


;ResetCmd:   cmpi.b       #esc,RecEsc    ;Vorig teken een "esc"?
;            bne          RecData2       ;Zo niet, dan teken alsnog in buffer zetten.
;            tst.b        ProtocolOn
;            beq          RecData2       ;Reset alleen als we met het Protocol werken
;            movem.l      (SP)+,A0/D0/D1 ;Haal reg van stack.
;            bra          Reset0         ;Geef reset.


*--------------------------------------
;TrData      tst          TrAantal      Zendbuffer leeg?
;            beq.s        TrNoData
;            lea          TrBuf,A0      Haal een teken uit de buffer.
;            move         TrLeesPtr,d0
;            move.b       0(A0,d0.w),AciaData Zend het teken.
;            addq         #1,d0         Leeswijzer 1 ophogen
;            andi         #TrBufLen-1,d0
;            move         d0,TrLeesPtr
;            subq         #1,TrAantal   Aantal tekens in buffer 1 aflagen
;            bra          AciaIntRet

;TrNoData    move.b       #AciaTrOff,(A0) Zet zendinterrupts uit.
;            bra          AciaIntRet

; --- ISR: ViaINT ------------------------------------------------------------
; 1991: Via-interruptroutine Timer-1 (100 Hz) en CA1 (1 Hz)
ViaInt:  	lea         ViaIFR,A0          ;IFR bevat interrupt status

;TIMER-1 (100 Hz) intterrupt afhandeling
            btst.b   	#IFR_TIMER1,(A0)   ;Test Timer 1 int(=free running/100HZ)
            beq.s	      ViaIntRTC          ;Geen interrupt Timer-1//test RTC

            bset.b      #IFR_TIMER1,(A0)   ;Clear int Timer-1 (door 1 te schrijven)
            addq.l      #1,TimerTicks      ;Hoog TimerTicks op
            tst.b       LEDfunction        ;Knipper LED=0?
            bne.s       ViaInt2

            move.l      TimerTicks,d0      ;Haal de tijd 
            and.w       #63,d0             ;640 ms
            bne.s       ViaInt2
            bchg.b      #PB2_LED,ViaPB     ;Blinck LED 640 ms
ViaInt2:    movem.l     (SP)+,d0/d1/a0     ;Herstel de registers
            move.l      FuzixIRQ,-(sp)     ;Zet adres Fuzix IRQ op stack
            rts                            ;Spring naar Fuzix, zodat FUZIX RTE kan doen!

 
;CA1/RTC (1Hz) interrupt afhandeling
ViaIntRTC:  btst.b      #IFR_CA1,(A0)      ;Test CA1-interrupt/1 sec RTC-tick?
            beq.s       ViaInt3

            bset        #IFR_CA1,(A0)      ;Clear CA1 (door 1 te schrijven!)
            addq.l      #1,TimeSec         ;Hoog de secondes op
            
ViaInt3:    movem.l     (SP)+,d0/d1/a0     ;Herstel de registers
NullProc:   rte


; === ISRjmptbl ==============================================================
; 2024-10-02 No spaces allowed around minus; calculation in vasm will go wrong!
ISRjmptbl:  dc.w        ISRmodem-ISRjmptbl   ;0: ISR modem int
            dc.w        ISRtransmit-ISRjmptbl;1: ISR transmit int
            dc.w        ISRrx-ISRjmptbl      ;2: ISR receive int
            dc.w        ISRerror-ISRjmptbl   ;3: ISR receiver line status int
            dc.w        ISRnone-ISRjmptbl    ;4: none
            dc.w        ISRnone-ISRjmptbl    ;5: none
            dc.w        ISRrx-ISRjmptbl      ;6: ISR receive int
            dc.w        ISRnone-ISRjmptbl    ;7: none


; --- ISR16550 ---------------------------------------------------------------
; 10-06-2023 ISR for UART 16550
ISR16550:   if UART16550     
            movem.l     d0/d1/a0/a1,-(sp)    ;Push used registers
            lea         UART16550,a0         ;A0:=Base UART 16550 address
ISR1:       move.b      ISR(a0),d0           ;D0:=Interrupt status register (ISR)
            btst        #0,d0                ;Interrupts pending?
            bne.s       ISRrdy               ;Ready(=1) if no interrupts are pending

            and.w       #$0E,d0              ;D0:=ISR (bit 1 to 3) 
            lea         ISRjmptbl(PC),a1
            move.w      (a1,d0.w),d0         ;Relocate address from table
            adda.w      d0,a1                ;A1:=points to routine from jump table
            jsr         (a1)                 ;Jump to subroutine

            bra.s       ISR1                 ;Test for more pending interrupts
            
ISRrdy:     movem.l     (sp)+,d0/d1/a0/a1    ;Pop used registers
            bclr        #St1FuzixCall,Status1;Check and clear Fuzix bit
            beq.s       ISRrdy1
            
            move.l      FuzixIRQ,-(sp)       ;Push Fuzis-address to stack
            rts                              ;RTS to Fuzix; FUZIX will clean up stack (RTE)

            endif;
ISRrdy1:    rte


; --- ISRerror(3) ------------------------------------------------------------
ISRerror:   if UART16550
            move.b      LSR(a0),d0           ;Clear pending error int
            endif 
ISRtransmit:
ISRnone:    rts
 
 
; --- ISRrx(2/6) -------------------------------------------------------------
ISRrx:      if UART16550
            move.b      RHR(a0),d0           ;Read 16550 FIFO to clear pending int
            cmpi.w      #RecBufLen,RecAantal ;Check for room
            bge.s       ISRrx1               ;No room in buffer-discard character!
            
            lea         RecBuf,a1            ;Get base address of receive buffer (a1)
            move.w      RecLeesPtr,d1        ;Get start buffer
            add.w       RecAantal,d1         ;Calculate write pointer end of buffer
            andi.w      #RecBufLen-1,d1      ;Circular buffer
            move.b      d0,0(a1,d1.w)        ;Write character into the buffer
            addq.w      #1,RecAantal         ;Increment number of characters in buffer

ISRrx1:     btst        #0,LSR(a0)           ;Another character available?
            bne.s       ISRrx
            endif
ISRrxRdy:   rts

  
; --- ISRmodem(0) ------------------------------------------------------------
; Check for 100Hz tick (Ring indicator)
;XXXXX BIJ SPRONG NAAR FUZIX HIER BIT OPZETTEN EN IN HOOFDROUTINE SPRINGEN
;XXXXX ZORG DAN DAT DE STACK SCHOON IS.
ISRmodem:   if UART16550
            move.b      MSR(a0),d0           ;D0:=Read Modem Status Register
            btst        #2,d0                ;Check for ring-indicator
            beq.s       ISRmodem1            ;Skip by other events

            bset        #St1FuzixCall,Status1;Set status for jump to fuzix
            addq.l      #1,TimerTicks        ;TimerTick 100Hz = 10 ms
            move.l      TimerTicks,d0
            and.w       #127,d0               ;Blink LED every 1280 ms
            bne.s       ISRmodem1
            bchg        #MCR_LED,MCR(a0)     ;Blink LED
            endif
ISRmodem1:  rts            

  
; === WaitTrEmpty ============================================================
; 05-02-2022 Wacht totdat de seriele verzendbuffer (UART) leeg is
;WaitTrEmpty:move.l       d0,-(sp)            ;Bewaar D0 op stack
;            moveq.l      #-1,d0              ;Maximaal 65.000 keer proberen
;WaitTrEmpty1:
;            tst.w        TrAantal            ;Wacht totdat zendbuffer leeg is
;            dbeq         d0,WaitTrEmpty1
;            move.l       (sp)+,d0
;            rts
 
 
            
; === DecOutD5 ===============================================================
; 26-11-2021 Zet D5.L op het scherm (SU-nodig)
; D2 geeft aan of er al output is geweest ('0') of niet ($0)
; A0: pointer naar tabel met decimalen; D0.l=decimaal
; D3: teller/digit
DecOutD5:   movem.l  d0-d3/a0,-(sp) ;Bewaar registers
            tst.l	   d5		         ;Test of D5.l negatief is
            bpl.s	   DecOut1
            move	   #'-',d1		   ;Print '-'
            moveq	   #6,d0
            trap	   #15
            neg.l	   d5		         ;Maak D5.l positief

DecOut1:    clr.l	   d2		         ;Initiatie; geen output geweest
            lea	   DecimalTbl(pc),a0 ;Tabel met tientallen
DecOut2:    move.l	(a0)+,d0		   ;Haal tiental op (100,10 of 1)
            beq.s	   DecOut5	      ;KLAAR!

            moveq	   #-1,d3		   ;Maak teller -1
            cmpi.l	#1,d0		      ;Bij "1" altijd output geven
            bne.s	   DecOut3
            moveq	   #'0',d2		   ;Geef output
DecOut3:    add.l	   #1,d3		      ;Hoog digit alvast op
            sub.l	   d0,d5		      ;Trek tiental af van te printen getal
            bpl.s	   DecOut3  	   ;Ga door zolang getal positief is (of 0)
            add.l	   d0,d5		      ;Hoog getal op (1 keer te vaak afgetrokken)

            tst.l	   d3		         ;Output aanzetten indien d3>0
            beq.s	   DecOut4
            moveq.l	#'0',d2		   ;Zet output aan; d2:='0'

DecOut4:    or.l	   d2,d3		      ;Trucje maak ASCII indien D3='0'
            beq.s	   DecOut2  	   ;Geen output, bereken volgende cijfer
            moveq	   #6,d0		      ;Print cijfer
            move.l	d3,d1
            trap	   #15
            bra.s	   DecOut2

DecOut5:    movem.l  (sp)+,d0-d3/a0 ;Herstel registers
            rts


DecimalTbl:	dc.l	   1000000000		;Tabel met decimalen ivm ontbreken DIV.L
            dc.l     100000000
            dc.l     10000000
            dc.l     1000000
            dc.l     100000
            dc.l     10000
            dc.l     1000
            dc.l     100
            dc.l     10
            dc.l	   1
            dc.l	   0              ;Einde van de tabel


; === PRIEMGETALLEN ==========================================================
; 1991: Zet de priemgetallen van 1..65000 op het scherm;
; Source gemaakt met de pascalcompiler.
; Roep procedure PROC4 aan.
Proc1:      link	      A6,#-6      ;PROGRAM PRIEMGRTAL ; FUNCTION PRIEM ( P : INTEGE
            move.w	   8(A6),D5	   ;BEGIN OK := ODD ( P )
            lsr.w	      #1,D5
            Scs	      -1(A6)

            move.w	   #3,-4(A6)	;D := 3

            move        Wortel,D5   ;max:=sqrt(p)
sqrt0:      move        d5,d4       ;while max*max<p DO inc(max)
            mulu        d4,d4
            cmp         8(a6),d4
            bhi.s       sqrt1
            addq        #1,d5
            bra.s       sqrt0
sqrt1:      move        d5,-6(A6)
            move        d5,Wortel

whle2:      move.b	   -1(A6),D5	;WHILE OK AND
            beq.s	      Cand3
            move.w	   -4(A6),D5	;( D <= MAX
            cmp.w	      -6(A6),D5
            Sls	      D5	)
Cand3:      tst.b       d5
            beq	      od2
            clr.l       d5
            move.w	   8(A6),D5	DO ;BEGIN OK := P MOD D
            move.w	   -4(A6),D4
            divu	      D4,D5
            clr.w	      D5
            swap	      D5
            Sne	      -1(A6)	   ;<> 0

            addq.w	   #2,-4(A6)	; INC ( D , 2 )
            bra      	whle2	END

od2:	      move.b	   -1(A6),10(A6)	; PRIEM := OK
            unlk	      A6	END
            rts

PrimeTxt:   dc.b        cr,lf,'Primes from 3 to 65000:',cr,lf,0
            align       2              ;Align 32 bits
            
MenuP:      link      A6,#-2	      ; PROCEDURE MAIN ; VAR P : INTEGER ;
            clr.w     Wortel
            clr.w     SLinenr       ;Teller aantal gevonden priemgetallen
            lea       PrimeTxt(PC),a1
            bsr       StrOutT15

            move.w	   #3,-2(A6)	   ;BEGIN FOR P := 1
            move.w	   #65000,-(SP)   ;TO MaxInt DO
            move.w	   -2(A6),D5
            cmp.w	      (SP),D5
            bhi	      efor5

body5:      subq.l	   #2,SP	         ;IF PRIEM
            move.w	   -2(A6),-(SP)	;( P
            bsr	      Proc1	         ;)
            addq.l	   #2,SP
            move.b	   (SP)+,D5
            beq	      else6
            clr.l       d5

            clr.l       d1
            move.w	   -2(A6),D1	   ;THEN WRITE ( P
            moveq       #3,d0          ;04-12-2021 Print priemgetal via TRAP
            trap        #15

            moveq       #' ',d1		   ;Print spatie via TRAP-15
            bsr         CharOutT15

;20-05-2024 CrLf for minicom after 12 primes            
            addq.w      #1,SLinenr        ;Hoog aantal gevonden priemgetallen op
            move.w      SLinenr,d0
            divu.w       #12,d0        ;Deelbaar door 12?
            swap        d0
            tst.w       d0
            bne         else6         
            bsr         T15CrLf        ;Print Cr&LF
            
else6:      move.w	   -2(A6),D5	   ;)
            cmp.w	   (SP),D5
            bhs.s	   efor5
            addq.w	   #1,-2(A6)
            bra	      body5
efor5:      addq.l	   #2,SP

            unlk	A6	; END
            trap     #0                ;Einde programma


; === Benchmarks =============================================================
; 2024-04-02 Jumptable and data for execution of the benchmarks
; 2024-10-02 No spaces allowed around minus; calculation in vasm will go wrong!

BenchTbl:   dc.l     Fib16Txt-BenchTbl       ;( 0) Benchmark text
            dc.l     Fibonacci16-BenchTbl    ;( 4) Startaddress FIB16( 24 )
            dc.l     10                      ;( 8) Repeat 10x
            dc.l     24                      ;(12) Parameter (d1)
            
            dc.l     Fib32Txt-BenchTbl       ;( 0) Benchmark text
            dc.l     Fibonacci32-BenchTbl    ;( 4) Startaddress FIB32( 24 )
            dc.l     10                      ;( 8) Repeat 10x
            dc.l     24                      ;(12) Parameter (d1)

            dc.l     SieveTxt-BenchTbl       ;( 0) Benchmark text
            dc.l     Sieve-BenchTbl          ;( 4) Startadres FIB16()
            dc.l     100                     ;( 8) Repeat 100x
            dc.l     24                      ;(12) Parameter (d1)

            dc.l     RamCopyByteTxt-BenchTbl ;( 0) Benchmark text
            dc.l     RamCopyByte-BenchTbl    ;( 4) Startasdress benchmark
            dc.l     1000                    ;( 8) Repeat 1000x
            dc.l     4096                    ;(12) Parameter (d1)=4kB

            dc.l     RamCopyWordTxt-BenchTbl ;( 0) Benchmark text
            dc.l     RamCopyWord-BenchTbl    ;( 4) Startaddress benchmark
            dc.l     1000                    ;( 8) Repeat 1000x
            dc.l     4096                    ;(12) Parameter (d1)=4kB

            dc.l     RamCopyLongTxt-BenchTbl ;( 0) Benchmark text
            dc.l     RamCopyLong-BenchTbl    ;( 4) Startaddress benchmark
            dc.l     1000                    ;( 8) Repeat 1000x
            dc.l     4096                    ;(12) Parameter (d1)=4kB

            dc.l     BitToggleTxt-BenchTbl   ;( 0) Benchmark tekst
            dc.l     BitToggle-BenchTbl      ;( 4) Startadres benchmark
            dc.l     1                       ;( 8) 1x testen
            dc.l     100                     ;(12) D1=100 ticks (=1 sec) testen

            dc.l     0                       ;Eind of table


; --- Benchmarks -------------------------------------------------------------
; 02-04-2024 Run a number of benchmarks by a table
MenuF:      lea      BenchTbl(pc),a0   ;A0:=jump table address
            move.l   a0,BenchmarkPtr   ;Save to memory
Benchmark1: movea.l  BenchmarkPtr,a0   ;Get the current table position
            movea.l  0(a0),a1          ;Get the benchmark text into A1
            tst.l    0(a0)             ;Ready?
            beq.s    Benchmark9

            lea      BenchTbl(PC),a2   ;Make text (A1) reloc
            add.l    a2,a1

            moveq    #T15StrOut,d0     ;Print benchmark text
            trap     #15
                        
            moveq    #T15GetTime,d0    ;Fetch time in 1/100 sec
            trap     #15
            move.l   d1,-(sp)          ;Save time to stack

            move.l   8(a0),-(sp)       ;Save nr of itterations to stack
Benchmark2: movea.l  BenchmarkPtr,a0   ;Get current table position
            move.l   12(a0),d1         ;Get parameter from table
            movea.l  4(a0),a0          ;Get benchmark start address from table
            lea      BenchTbl(PC),a2   ;Make address (a0) reloc
            add.l    a2,a0
            jsr      (a0)              ;Execute benchmark            

            subq.l   #1,(sp)           ;Itt := Itt -1
            bne.s    Benchmark2        ;Repeaut until Itt=0
            addq.l   #4,sp             ;Clean stack from itt

            move.l   d0,d1             ;Print onces the result of Fib(n)
            beq.s    Benchmark3        ;Only if result not zero
            moveq    #T15DecOut,d0
            trap     #15

Benchmark3: lea      BenchmarkReadyTxt(pc),a1  ;Print "Ready in "
            moveq    #T15StrOut,d0   
            trap     #15

;XXXX GEBRUIK LEAPTIME SUBROUTINES XXXX
            moveq    #T15GetTime,d0    ;Fetch time in 1/100 sec
            trap     #15
            sub.l    (sp)+,d1          ;Calculate leaptime in 1/100 sec
            mulu.w   #10,d1            ;Calculate mSec
            
            moveq    #T15DecOut,d0     ;Print leaptime
            trap     #15
  
            lea      BenchmarkMsTxt(pc),a1
            move     #T15StrOut,d0     ;Print "mSec"
            trap     #15

            add.l    #16,BenchmarkPtr  ;Go to next benchmark in table
            bra      Benchmark1
    
Benchmark9: rts                        ;Return to BIOS-menu
            


; --- Fibonacci32 ------------------------------------------------------------
; 26-11-20221 Calculate Fibonacci (32 bits)
; D0.l = output
; D1.l = input
; Fib(1)=1, Fib(2)=1, Fib(n)=Fib(n-1) + Fib(n-2)
Fibonacci32:moveq.l  #1,d0       ;Fib(1)=1, Fib(2)=1
            cmpi.l   #2,d1        
            ble.s    Fibonacci32A;Fib(1)=1 en Fib(2)=2

            subq.l   #1,d1       ;D1 := n-1
            move.l   d1,-(sp)    ;Save register D1=n-1
            bsr      Fibonacci32 ;D0 := Fib(n-1)
            move.l   (sp)+,d1    ;Restore D1 = n-1
            move.l   d0,-(sp)    ;Save Fib(n-1) to stack

            subq.l   #1,d1       ;D1 = n-2
            bsr      Fibonacci32 ;D0 := Fib(n-2)
            add.l    (sp)+,d0    ;D0 := Fib(n-1) + fib(n-2)

Fibonacci32A:rts
 

; --- Fibonacci16 ------------------------------------------------------------
; 26-11-20221 Calculate Fibonacci (32 bits)
; D0.w = output
; D1.w = input
; Fib(1)=1, Fib(2)=1, Fib(n)=Fib(n-1) + Fib(n-2)
Fibonacci16:moveq.l  #1,d0       ;Fib(1)=1, Fib(2)=1
            cmpi.w   #2,d1        
            ble.s    Fibonacci16A;Fib(1)=1 en Fib(2)=2

            subq.w   #1,d1       ;D1 := n-1
            move.w   d1,-(sp)    ;Save register D1=n-1
            bsr      Fibonacci16 ;D0 := Fib(n-1)
            move.w   (sp)+,d1    ;Restore D1 = n-1
            move.w   d0,-(sp)    ;Save Fib(n-1) to stack

            subq.w   #1,d1       ;D1 = n-2
            bsr      Fibonacci16 ;D0 := Fib(n-2)
            add.w    (sp)+,d0    ;D0 := Fib(n-1) + fib(n-2)

Fibonacci16A:rts


; --- SIEVE ------------------------------------------------------------------
; 00-00-2023 Calculate primes using a array of 8192 bytes
; OUT: d0 = found primes 
; A0=start array
; D0=index, D1=prime/index delete, D2=nr found primes
SieveSize:  equ      8190              ;Nr of bytes used (must be even)

Sieve:      sub      #SieveSize+2,SP   ;Allocatie space on stack (use LNK instead?)
            lea      (sp),a0           ;A0 start array

            move.w   #SieveSize-1,d0   ;Fill array with -1
            moveq.l  #-1,d1
Sieve1:     move.b   d1,(a0)+          ;Fill with -1
            dbf      d0,Sieve1
            
            lea      (sp),a0           ;Start array
            clr.l    d0                ;Array pointer (d0)
            clr.l    d2                ;nr of primes (d2)
Sieve2:     tst.b    0(a0,d0.l)        ;Test if prime?
            beq.s    Sieve4            ;Jump if no prime

            addq.l   #1,d2             ;Increment nr of primes found
            move.l   d0,d1             ;Calculate prime (=d0*2 + 3)
            add.l    d1,d1             ;d1:=d1*2
            addq.l   #3,d1             ;d1:=d1*2 + 3 (=prime)

;            movem.l  d0/d1,-(sp)       ;Push registers
;            moveq    #T15DecOut,d0     ;Print prime (d1)
;            trap     #15
;            movem.l  (sp)+,d0/d1       ;pop registers
            
            move.l   d0,d3             ;D3=index
Sieve3:     add.l    d1,d3             ;Delete all ascenders from array
            cmp.l    #SieveSize,d3     ;End of array/ready?
            bge.s    Sieve4
            clr.b    0(a0,d3.l)        ;****Clear multiple
            bra.s    Sieve3
            
Sieve4:     addq.l   #1,d0             ;Next prime/index
            cmp.l    #SieveSize,d0     ;Ready with all the primes?
            blt.s    Sieve2
                       
            add      #SieveSize+2,SP   ;Release stack space
            move.l   d2,d0             ;Return nr of found primes
            rts


; === RamCopyByte ============================================================
; 03-04-2024 Copy D1.w bytes from EPROM to RAM
; D1.w = nr of bytes to copy (must be even). Copy BYTES
RamCopyByte:suba.l   d1,SP             ;Create space to copy to
            move.l   d1,d0             ;D0=byte counter
            subq.l   #1,d0             ;DBF correction
            
            lea      RomVector_00(pc),a0  ;Copy from A0 (EEPROM)
            movea.l  SP,a1             ;... to A1 (=SP) 
RamCopyB1:  move.b   (a0)+,(a1)+       ;move (8 bits)
            dbf      d0,RamCopyB1      ;loop
            adda.l   d1,SP             ;Release stack space
            clr.l    d0                ;Return zero
            rts
 

; --- RamCopyWord ------------------------------------------------------------
; 03-04-2024 Copy D1.w bytes from EPROM to RAM
; D1.w = nr of bytes to copy (must be even). Copy WORDS (16 bits)
RamCopyWord:suba.l   d1,SP             ;Create space to copy to
            move.l   d1,d0             ;D0=byte counter
            asr.l    #1,d0             ;Divide D0 by two
            subq.l   #1,d0             ;DBF correction
            
            lea      RomVector_00(pc),a0  ;Copy from A0 (EEPROM)
            movea.l  SP,a1             ;... to A1 (=SP) 
RamCopyW1:  move.w   (a0)+,(a1)+       ;move word (16 bits)
            dbf      d0,RamCopyW1      ;loop
            adda.l   d1,SP             ;Release stack space
            clr.l    d0                ;Return zero
            rts
 

; --- RamCopyLong ------------------------------------------------------------
; 03-04-2024 Copy D1.w btes from EPROM to RAM
; D1.w = nr of bytes to copy (must be even). Copy LONGWORDS (32 btis)
RamCopyLong:suba.l   d1,SP             ;Create space to copy to
            move.l   d1,d0             ;D0=byte counter
            asr.l    #2,d0             ;Divide D0 by two
            subq.l   #1,d0             ;DBF correction
            
            lea      RomVector_00(pc),a0     ;Copy from A0 (EEPROM)
            movea.l  SP,a1             ;... to A1 (=SP) 
RamCopyL1:  move.l   (a0)+,(a1)+       ;move longword (32 bits)
            dbf      d0,RamCopyL1      ;loop
            adda.l   d1,SP             ;Release stack space
            clr.l    d0                ;Return zero
            rts


; --- BitToggle ------------------------------------------------------------
; D0=Time, input: D1.L=ticks, D2=counter, D3=1000x, A0=IO chip
BitToggle:  move.l   TimerTicks,d0     ;Get time/test for timer
            beq.s    BitToggle9        ;No timer/no test
            add.l    d1,d0             ;Calculate stop time (D0)

            trap     #1                ;Become SU-to directly toggle IO
            clr.l    D2                ;Init registers; clear counter(D2)
BitToggle2: move.w   #999,d3           ;Toggle 1000x (D3)

            if UART16550
            lea      UART16550+MCR,A0  ;IO address 16550 UART
            else
            lea      ViaPB,A0          ;IO address 6522 VIA
            endif

BitToggle3: if UART16550
            bset     #MCR_LED,(A0)     ;LED on
            bclr     #MCR_LED,(A0)     ;LED off
            else                       ;VIA 6522
            bclr.b   #PB2_LED,(A0)     ;LED on 
            bset.b   #PB2_LED,(A0)     ;LED off 
            endif            
            dbf      d3,BitToggle3

            addq.l   #1,D2             ;Increment counter
            cmp.l    TimerTicks,d0     ;ready?
            bgt.s    BitToggle2
            
            divu.w   #100,d1           ;100 ticks=1 sec=1 kHz
            mulu     d1,d2             ;Vermenigvuldig met gemeten doorloop
            clr      d0
            move.w   d2,d0             ;Return kHz in D0.L
            
            move.w   #$0000,SR         ;Become normal user again
BitToggle9: rts
 

 
; === Memory hexdump =========================================================
; 25-05-2024 Memory hexdump + ASCII
MenuH:
Hexdump:    moveq       #14,d0          ;Print "Address:"
            lea         HexdumpAddrTxt(pc),a1
            bsr         StrOutT15
            
            moveq       #2,d0          ;D1=command 2: Read line (A1, null terminated)
            lea         SLineBuff,A1   ;A1=Buffer
            trap        #15

            bclr        #St1SRecError,Status1 ;Clear error flag
            lea         SLineBuff,A1   ;A1=Buffer
            moveq.l     #0,d0          ;Initialize D0
Hexdump0:   tst.b       (a1)           ;Empty line/ready?
            beq         Hexdump1       ;Start dump with currect address
            moveq.l     #0,d1          
            bsr         SReadNibble    ;Read nibble(1 byte), d0=0
            bne         Menu           ;No hex; return to Menu
            move.l      d0,HexdumpAddr ;Save correct dumpaddress
            bra         Hexdump0       ;Read next nibble

Hexdump1:   moveq.l     #15,d3         ;Print 16 rows at a time
Hexdump2:   bsr         T15CrLf
            move.l      HexdumpAddr,d0 ;Print address hex
            andi.l      #$FFFFFFF0,d0  ;Alwasy start at 16 bytes
            move.l      d0,HexdumpAddr ;Save (altered) address

            bsr         HexOut         ;Print hex address
            lea         HexdumpColonTxt(pc),a1 ;Print ": "
            bsr         StrOutT15
            
;Print 16x hex
            movea.l     HexdumpAddr,a0 ;Load dump address
            moveq       #7,d2          ;8 words in a row
Hexdump3:   move.w      (a0)+,d0
            bsr         HexWrdOut      ;Print memory/D0 hexadecimal
            moveq.l     #' ',d1
            bsr         CharOutT15     ;Print a space 

            move.l      a0,d0          ;Print extra space after 8 bytes
            andi.w      #7,d0
            bne.s       Hexdump4
            bsr         CharOutT15     ;Print space 

Hexdump4:   dbf         d2,Hexdump3    ;Process next memory word
                

;Print 16x ASCII
            movea.l     HexdumpAddr,a0 ;Load dump address
            moveq       #15,d2         ;16 chars in a row
Hexdump5:   moveq.l     #'.',d1        ;Prepair dot char in D1
            move.b      (a0)+,d0
            cmpi.b      #' ',d0
            blt         Hexdump6       ;Print ctrl char as dot(.)
            cmpi.b      #127,d0         
            bgt         Hexdump6       ;Print high chars as dot(.)
            move.b      d0,d1
Hexdump6:   bsr         CharOutT15     ;Print memory byte as char 

Hexdump7:   dbf         d2,Hexdump5
            move.l      a0,HexdumpAddr ;Save dumpaddress for next round
            dbf         d3,Hexdump2
            bra         Hexdump


; === S-RECORD LOADER ========================================================
; 02-01-2022 Read S-Record file by console/UART and start the program
; A1=Linebuffer, A2=Store data address, D6=offset, D7=checksum 
; Loading from 0x0000 means the program can placed anywhere in the memory and the  
; offset (D6) 0x10400 (Start free RAM)
; 09-08-2023 Integratie met XModem, tbv FLterm/minicom           
MenuS:     lea         SRecStartTxt(pc),a1
            bsr         StrOutT15      ;Print "Load S-records:"
            bsr         SRecInit       ;Init SRec/XModem variables

SReadLine:  clr.b       EchoInput      ;Suppress echo during receiving S-records
            moveq       #2,d0          ;D1=command 2: Read S-line (A1, null terminated)
            lea         SLineBuff,A1   ;A1=Buffer
            trap        #15
            st          EchoInput      ;Switch echo on

            bsr         SProcessLine   ;Process S-line
            move.l      LeapCounter,d1 ;Show kB's (if changed)
            asr.l       #8,d1          ;Calculate kB's, divide by 1024 (>>10)
            asr.l       #2,d1          
            cmp.w       SKBsize,d1     ;kB's Changed?
            beq.s       SReadLine1
            move.w      d1,SKBsize     ;Store new kBs
            
            moveq       #T15DecOut,d0  ;Print #kBs
            trap        #15
            lea         SRecKbTxt(pc),a1 ;Print " kB<cr>"
            moveq       #T15StrOut,d0  
            trap        #15

SReadLine1: btst        #St1SRecError,Status1;Check error flag
            bne         SReadLine2
   
            tst.l       SRecPrgStart   ;Start address program entered?
            beq.s       SReadLine      ;No, read next line
            bra         SPrgStart      ;Start program!
SReadLine2: rts                        ;Return after error
            

; --- SProcessLine -----------------------------------------------------------
; 09-08-2023 Subroutine verwerkt één ingelezen S-regel; ook voor XModem 
SProcessLine:
            lea         SLineBuff,A1   ;A1=begin ingelezen regel                   
            tst.b       (a1)           ;Lege regel? 
            beq.s       SRecRts        ;Klaar indien regel leeg is

            addq.w      #1,SLinenr     ;Hoog Regelnr op (niet bij lege regel)
            
            cmpi.b      #'S',(a1)+     ;Eerste teken moet "S" zijn
            bne.s       SRecFout
            
            move.b      (a1)+,d3       ;Volgend teken (1 t/m 9) in D3
            clr.l       d7             ;Clear checksum (D7)

            bsr         SReadByte      ;Skip next (length) byte; calc checksum
            bne.s       SRecFout
       
            clr.l       d0             ;D0:=0 (for Hex>binary SRecReadxBytes)
            cmpi.b      #'3',d3        ;S3=data (32 bits adres)
            beq.s       SRecS3Data
            cmpi.b      #'2',d3        ;S2=data (24 bits adres)
            beq.s       SRecS2Data
            cmpi.b      #'1',d3        ;S1=data (16 bits adres)
            beq.s       SRecS1Data
            
            cmpi.b      #'9',d3        ;S9=start program (16 bits adres)
            beq.s       SRecS9
            cmpi.b      #'8',d3        ;S8=start program (24 bits adres)
            beq.s       SRecS8
            cmpi.b      #'7',d3        ;S7=start program (32 bits adres)
            beq.s       SRecS7         ;Ander type record; gewoon overslaan

SRecRts:    rts                        ;Klaar           
            
            
; --- SChkFout/SRecFout ------------------------------------------------------
; 02-01-2022 Print diverse S-record foutmeldingen
SChkFout:   lea         SChkFoutTxt(pc),a1   ;Melding "Wrong checksum in line "
            bra.s       SRecFout1

SRecFout:   lea         SRecFoutTxt(pc),a1   ;Melding "Wrong format in line "
SRecFout1:  bsr         EmptyBuffer          ;Remove pending characters (in buffer)
            bset        #St1SRecError,Status1;Set error flag (for later use)
            bsr         StrOutT15            ;Print S-record error message

            clr.l       d1                   ;Toon regelnummer + CrLf
            move.w      SLinenr,d1
            moveq       #T15DecOut,d0
            trap        #15
            bsr         T15CrLf              ;Cr&Lf 

            lea         SLineBuff,A1         ;A1=begin ingelezen regel  
            bra         StrOutT15            ;Retour aanroeper (MENU)


; --- SRecS7/8/9 -------------------------------------------------------------
; 02-01-2022 Start programma via S7/S8/S9
SRecS7:     bsr         SReadByte      ;S7=start programma; lees 32 bits adres
SRecS8:     bsr         SReadByte1     ;S8=start programma; lees 24 bits adres
SRecS9:     bsr         SReadByte1     ;S9=start programma; lees 16 bits adres
            bsr         SReadByte1
SRecStart:  bne         SRecFout       ;Abort if errorflag set
            add.l       SRecOffset,d0  ;Tel de offset erbij
            move.l      d0,SRecPrgStart;Bewaar programma start 
            
            bsr         SReadByte      ;Lees checksum
            cmpi.b      #$FF,d7        ;Controleer checksum
            bne         SChkFout

            rts                        ;Ga terug naar hoofdprogramma (XModem)
            
            
; --- S19ProcessData ---------------------------------------------------------
; 02-01-2022 Subroutine verwerkt S1/S2/S3-data, schrijf in geheugen 
SRecS3Data: bsr         SReadByte     ;Lees 32 bits adres in D0
SRecS2Data: bsr         SReadByte1    ;Lees 24 bits adres in D0
SRecS1Data: bsr         SReadByte1    ;Lees 16 bits adres in D0
            bsr         SReadByte1
SRecData:   bne         SRecFout       ;Abort if error (NE)

            tst.l       SRecOffset     ;Test of offset al geïniteerd is (-1)
            bpl.s       SProcessData0
            tst.l       d0             ;Als D0=0 dan offset :=0x10400
            beq.s       SProcessData1
            clr.l       SRecOffset     ;Als D0<>0 dan offset :=0x0000
            bra.s       SProcessData0
SProcessData1:
            move.l      #Program,d6    ;Offset wordt 0x10400/0C00800
            if BIOSinRAM
            pea         ProgEnd(pc)    ;First free byte after this program
            move.l      (sp)+,d6   
            endif
            move.l      d6,SRecOffset
            
SProcessData0:
            add.l       SRecOffset,d0  ;Bereken adres incl offset
            move.l      d0,a2          ;Bewaar d0 in A2 (store data adres)
SProcessData:
            bsr         SReadByte      ;Lees volgend data byte
            bne         SRecFout       ;Foute hex-data; geef foutmelding

            tst.b       (a1)           ;Einde regel?
            beq.s       SCheckCrc      ;Controleer dan checksum
            move.b      d0,(a2)+       ;Schrijf data in geheugen
            addq.l      #1,LeapCounter ;Tel aantal geschreven databytes 
            bra.s       SProcessData   ;Herhaal data tot einde regel
            
SCheckCrc:  cmpi.b      #$FF,d7        ;Checksum staat in D7
            bne         SChkFout       ;Foute checksum; geef melding
            rts                        ;Klaar
            
            
; --- SReadNibble/Byte/2Byte/3Byte -------------------------------------------
; 02-01-2022 Lees hex byte (A1) in D0; Bereken checksum D7. Foute invoer: N-flag
SRead4Bytes:bsr         SReadByte      ;Lees eerste byte (maak D0:=0)
            bsr         SReadByte1     ;Lees tweede byte
            bsr         SReadByte1     ;Lees derde byte
            bra.s       SReadByte1     ;Lees vierde byte + retour aanroeper

SRead3Bytes:bsr         SReadByte      ;Lees eerste byte (maak D0:=0)
            bsr         SReadByte1     ;Lees tweede byte
            bra.s       SReadByte1     ;Lees derde byte + retour aanroeper

SRead2Bytes:bsr         SReadByte      ;Lees eerste byte (maak D0:=0)
            bra.s       SReadByte1     ;Lees tweede byte+retour aanroeper

            
; --- SReadByte ------------------------------------------------------------            
; 02-01-2022 Lees Nibble (A1) in D0. Error: set errorflag
; Output: D0.
; Uses: D1: counter, D2
SReadByte:  clr.l       d0                ;D0 := 0 (beginwaarde)
SReadByte1: moveq.l     #1,d1             ;Read 2 nibbles for 1 byte
SReadNibble:move.b      (a1)+,d2          ;Lees hex-teken in D2
            beq.s       SReadNibbleFout   ;Test for end-of-line (#0)
            subi.b      #'0',d2           ;Fout indien kleiner dan "0"
            blt.s       SReadNibbleFout

            cmpi.b      #9,d2             ;Klaar indien 0..9
            ble.s       SReadNibble1
            subi.b      #7,d2             ;Converteer A..F naar 10..15
            blt.s       SReadNibbleFout   ;Foute invoer
            cmpi.b      #15,d2
            bgt.s       SReadNibbleFout   ;Foute invoer
            
SReadNibble1:lsl.l      #4,d0             ;D0 := D0 * 16
            add.b       d2,d0             ;D0 := D0 + D2 (=hexwaarde)
SReadNibble2:
            dbf         d1,SReadNibble    ;Read next nibble
            add.b       d0,d7             ;Calculate byte checksum
            btst        #St1SRecError,Status1;Set/clear error flag (Z)
            rts

SReadNibbleFout:
            bset        #St1SRecError,Status1
            btst        #St1SRecError,Status1   ;Set NZ-flag
            rts
            
            
; --- StartProgram -----------------------------------------------------------
; Start program in user modes
; 2022-01-30 Voorlopig tbv FUZIX: wordt SU1
MenuR:
SPrgStart:  bsr         EmptyBuffer    ;Leeg UART/inputbuffer
            tst.l       SRecPrgStart   ;Test if program is loaded
            beq         SPrgError
            
;Clear instruction cache (68020) without changing te cache status (enabled/disabled)             
            move        #T15CacheCtrl,d0
            moveq.l     #-1,d1         ;Query cache status
            trap        #15
            ori.w       #8,d1          ;Set clear cache flag
            move        #T15CacheCtrl,d0
            trap        #15
            
            ;trap        #1             ;Wordt SU (FUZIX draait in SU)
            movea.l     MaxMem,SP      ;(S)SP:=Hoogste memory adres
            
            pea         ReadyProg(pc)  ;Return naar TRAP #0 naar stack (tbv RTS)
            move.l      SRecPrgStart,-(sp);Zet het startadres programma op stack
            
            lea         SRecRunTxt(pc),a1 ;Print "Start program at 0x"
            bsr         StrOutT15
            
            move.l      (sp),d0        ;Print startadres, maar laat het op stack staan!
            bsr         HexAddrOut
            bsr         T15CrLf        ;Print Cr&Lf
            

; 2022-01-30 Voorlopig tbv FUZIX: wacht totdat seriële buffers leeg zijn
;            bsr         WaitTrEmpty    ;Wacht totdat de UART-verzendbuffer leeg is
                     
            clr.l        D0            ;Maak alle registers nul
            clr.l        D1
            clr.l        D2
            clr.l        D3
            clr.l        D4
            clr.l        D5
            clr.l        D6
            clr.l        D7
            movea.l      D0,A0
            movea.l      D0,A1
            movea.l      D0,A2
            movea.l      D0,A3
            movea.l      D0,A4
            movea.l      D0,A5
            movea.l      D0,A6

            rts                        ;Start adres programma staat op stack


SPrgError:  lea         SPrgErrorTxt(pc),a1; Show txt: No program loaded
            bsr         StrOutT15
            bra         Menu           ;Return to Menu   

ReadyProg:  trap         #0            ;Geef na afloop userprogramma een trap
                                       ;om weer supervisor te worden.


;--- SRecInit ----------------------------------------------------------------
; 09-08-2023 SRec en XModem werken samen; één initiatie routine
SRecInit:   clr.w       SLinenr        ;Linecount := 0
            clr.w       SKBsize        ;Size      := 0 kB
            bsr         InitLeapTime   ;Init leaptime/counter (kB-teller)
            moveq       #-1,d6         ;Initieer offset met -1
            move.l      d6,SRecOffset  ;Init offset met -1
            clr.l       SRecPrgStart   ;Init start adsress program
            bclr        #St1SRecError,Status1; Clear error flag
            
            move.l      #Program,HexdumpAddr ;File init hexdump address
            rts
            
            
;=== XModem ==================================================================
;18-06-2023 Vereenvoudigde XModem-CRC ontvanger; bij fouten direct stoppen
;FLTerm kent alleen de klasieke XModem met 8 bits XOR.
MenuX:
XModem:     lea         XModemStartTxt(pc),a1 ;Zet XModem starttekst op scherm
            bsr         StrOutT15
			
            move.b      ProtocolOn,d0        ;Bewaar protocol status
            move.b      d0,XModemProto
            clr.b       ProtocolOn           ;Zet Xon/Xoff uit!!!
            
            lea		   SLineBuff,a0
            move.l	   a0,XModemSLine		   ;Initieer pointer SLine
            move.w      #1,XModemBloknr      ;Initieer eerste bloknr = 1
            
XModemReceive:
            move.w      #60,d4               ;Initieer timers (60x proberen)
XModem0:    moveq       #-1,d3
XModem1:    lea         XModemNAKtxt(pc),a1  ;Zend NAK (start protocol) of ACK
            cmpi.w      #1,XModemBloknr      ;Bloknr 1->NAK, ander zend ACK
            beq.s       XModem1b	
            lea         XModemACKtxt(pc),a1  ;Zend ACK (blok goed ontvangen)
XModem1b:   bsr         StrOutT15
       
XModem2:    moveq       #T15KeyPressed,d0    ;Teken ontvangen vanuit zender?
;XXXX Tijdelijk springen naar routine, totdat BIOS/TRAP15 geüpdate is
;            trap        #15           
 bsr Trp15CharAvail

            tst.b       d1
            beq.s       XModem3              ;Geen teken; laag teller af
            bsr         CharInT15            ;Lees teken, moet SOH, CAN of EOT zijn
            cmpi.b      #SOH,d1              ;SOH=start 1e blok
            beq         XModemSOH              

            lea         XModemCancelTxt(pc),a1
            cmpi.b      #CAN,d1              ;CAN?
            beq         XModemError          ;Error: CANCEL-melding+stoppen

            cmp.b       #EOT,d1              ;EOT=transfer is klaar
            beq         XModemKlaar

XModem3:	   dbf			d3,XModem2			 ;Ander teken, dan leesbuffer verder testen
            dbf			d4,XModem1			 ;Stuur opnieuw NAK
            lea         XModemTimeoutTxt(pc),a1 
            bra			XModemError          ;Error: timeout (geen SOH)


;SOH is ontvangen; start nieuw blok
XModemSOH:  bsr         CharInT15            ;Lees bloknummer (D1.b)
            lea         XModemBloknrfoutTxt(pc),a1 ;Tekst bloknr fout
            move.w		XModemBloknr,d0		 ;Te verwachten bloknr in D0.w
            cmp.b       d0,d1      			 ;Controleer bloknr met verwacht (byte!)
            bne         XModemError          ;Error: bloknr fout

            bsr         CharInT15            ;Lees geïnverteerde bloknummer
            not.b       d1                   ;Inverteer
            cmp.b       d0,d1			         ;Controleer met wat we verwachten
            bne         XModemError

            clr.w       XModemCRC            ;Initieer CRC/checksum op nul
            move.w      #128,XModemTeller    ;Bloksize = 128
XModemBlok: bsr         CharInT15            ;Lees 1 teken in D1
            add.b       d1,XModemCRC         ;Werk de checksum bij

            movea.l     XModemSLine,a0       ;Schrijfpointer
            cmpi.b      #' ',d1              ;Cr, Lf of ˆZ is einde regel
            blt			XModemRegeleinde
            move.b		d1,(a0)+             ;Schrijf teken in de buffer
            move.l      a0,XModemSLine       ;Bewaar schrijfpointer

XModemBlok1:subq.w      #1,XModemTeller
            bne.s       XModemBlok

            bsr         CharInT15            ;Lees Checksum/CRC
            lea         XModemCRCtxt(pc),a1  ;Melding: checksum fout+CAN
            cmp.b       XModemCRC,d1         ;Controleer de CRC
            bne         XModemError          ;Error: checksum fout

;Eén blok correct gelezen; hoog bloknr op en stuur [ACK]
            addq.w      #1,XModemBloknr      ;Hoog bloknummer op (w)
            bra         XModemReceive        ;Lees volgend blok


;--- Regeleinde: 1 regel gelezen ---
XModemRegeleinde:
            clr.b       (a0)                 ;Sluit de string af met een null
            lea         SLineBuff,a0         ;Begin van S-regelbuffer
            move.l      a0,XModemSLine       ;Initieer alvast de volgende regel
;XXXXX IS TEST OP NUL NODIG OF GEBEURT DAT IN SREC?
;            tst.b       (a0)                 ;Test op lege string
;            beq         XModemBlok1          ;Lees blok verder in

;            lea         XModemSRECtxt(pc),a1  ;Melding: S-record fout
;            cmpi.b      #'S',(a0)            ;Regel moet beginnen met een S
;            bne         XModemError          ;Error: No S-line
;XXXX VERWERK REGEL
;			bsr			SRecLine             ;XXXXX VERWERK S-REGEL XXXX

            bra         XModemBlok1          ;Lees blok verder in


;-- XModemKlaar: einde transfer ---         
XModemKlaar:            
            move.b      XModemProto,d0       ;Zet protocolstatus terug
            move.b      d0,ProtocolOn

            lea         XModemKlaarTxt(pc),a1 ;ACK+Transfer completed op scherm
            bsr         StrOutT15
;XXXXX Start HET PROGRAMMA XXXX
            trap        #0					;Terug naar BIOS

           
;Geef foutmelding (a1) met CAN en ga terug naar BIOS 		    
XModemError:move.l      a1,-(sp)       ;Geef eerst 2x Cancel
            lea         XModemCANtxt(pc),a1
            bsr         StrOutT15


;Herhaal de foutmelding met tussenposen omdat Kea die blijft opeten
;XXXXX EIGENLIJK HIER LEESBUFFER LEGEN (DAT WACHT OOK MAAR IS NU STUK)            
XModemErr0: move.l      #500000,d0
XModemErr1: subq.l      #1,d0
            bne.s       XModemErr1     ;Wacht daarna een tijdje

            move.l      (sp)+,a1       ;Geef de echte foutmeldingstekst
            bsr         StrOutT15
            bra.s       XModemError    ;Blijf foutmelding herhalen
            
            move.b      XModemProto,d0       ;Zet protocolstatus terug
            move.b      d0,ProtocolOn

            trap        #0             ;Terug naar menu

            
; --- CRC16 ------------------------------------------------------------------
; Bereken on the fly CRC16 voor XModem (poly=0x1021, start=0x000)
; Input:   D0.b nieuwe data (byte) waarde
; Uit:     XModemCRC = CRC16 waarde
; Gebuikt: D0 t/m D1
;CRC16:      lsl.w       #8,d0                ;Input in hoge bits D0 [15-8]
;            move.w      XModemCRC,d1         
;            eor.w       d1,d0                ;crc := crc xor data<<8
;            moveq       #7,d1                ;Doorloop de lus 8x
;CRC16_0:    lsl.w       #1,d0                ;crc := crc << 1
;            bcc.s       CRC16_1              ;if crc & 0x1.0000
;            eor.w       #XModemPoly,d0       ;crc := crc xor poly
;CRC16_1:    dbra        d1,CRC16_0           ;doe dit 8x        
;            move.w      d0,XModemCRC         ;Bewaar resultaat
;            rts            


; === T15CrLf ================================================================
; Toon Cr&Lf via Trap-15 (usermodes)
T15CrLf:    movem.l     d0/a1,-(sp)          ;Bewaar registers
            lea         CrLfTxt(pc),a1       ;Toon CrLf
            moveq       #T15StrOut,d0
            trap        #15

            movem.l     (sp)+,d0/a1          ;Red de registers
            rts                              ;Terug naar aanroeper


; === EmptyBuffer ============================================================
; 02-01-2022 Leeg de RS232 keyboard/inputbuffers na een fout
EmptyBuffer:movem.l     d1/d2,-(sp)          ;Bewaar register d1/d2
            bra.s       EmptyBuf2
EmptyBuf1:  tst.w       RecAantal            ;Aantal tekens in buffer
            beq.s       EmptyBuf3            ;Geen teken; laag teller af
            bsr         CharInT15            ;Lees teken uit buffer
EmptyBuf2:  moveq.l     #-1,d2            ;Herlaad teller opnieuw
EmptyBuf3:  dbra        d2,EmptyBuf1
   
            movem.l     (sp)+,d1/d2          ;Red de gerbuikte registers
            rts                              ;Terug naar aanroeper
 
 
; --- InitLeapTime -----------------------------------------------------------
; Initieer LeapTime counters
InitLeapTime:clr.l      LeapCounter          ;LeapCounter := 0 (=aantal bytes)
            move.l      TimerTicks,LeapTime  ;Init leaptime met TimerTicks
            rts

IncLeapCounter:
            addq.l      #1,LeapCounter       ;Hoog leapcounter met één ophogen
            rts
            
ShowLeapTime:
            movem.l     d0/d1/a1,-(sp)       ;Bewaar registers
            moveq       #T15DecOut,d0        ;Toon <LeapCounter>
            move.l      LeapCounter,d1
            trap        #15

            moveq       #T15StrOut,d0
            lea         Leap1Txt(pc),a1      ;"...bytes loaded in..."
            trap        #15

            move.l      TimerTicks,d1        ;Haal huidige tijd
            sub.l       LeapTime,d1          ;Bereken leaptime (10ms)
            move.w      d1,-(sp)             ;Bewaar verschiltijd; LET OP ALS WOORD!
            divu        #100,d1              ;Bereken leaptime in sec
            andi.l      #$FFFF,d1            ;Maak hoge deel leeg (bevat na delen de rest)
            moveq       #T15DecOut,d0
            trap        #15                  ;<tijd in sec>
            
            moveq       #T15StrOut,d0
            lea         Leap2Txt(pc),a1      ;"...sec (..."
            trap        #15
            
            move.w      (sp)+,d0              ;Onderstaande kun je niet uitrekenen als tijd=0
            beq.s       ShowLeapTime1
            move.l      LeapCounter,d1       ;Haal aantal bytes
            mulu.w      #100,d1              ;D1:=Aantal Bytes * 100
            divu.w      d0,d1                ;D1:=Aantal bytes/sec
            andi.l      #$FFFF,d1            ;Maak hoge deel leeg
            moveq       #T15DecOut,d0        ;<bytes per sec>
            trap        #15
            
            moveq       #T15StrOut,d0
            lea         Leap3Txt(pc),a1      ;" bytes per sec)"
            trap        #15

ShowLeapTime1:            
            movem.l     (sp)+,d0/d1/a1       ;Red registers
            rts

            
; --- StrOutT15 --------------------------------------------------------------
; 09-01-2022 Print nul-terminated string A1 via trap-15 (user-mode)
StrOutT15:  move.l      d0,-(sp)             ;Bewaar D0
            moveq       #T15StrOut,d0        ;Print string A1
            trap        #15
            move.l      (sp)+,d0             ;Red D0
            rts
            
            
; --- CharInT15 --------------------------------------------------------------
; 09-01-2022 Input 1 teken in D1 via trap-15 (user-mode)
CharInT15:  move.l      d0,-(sp)             ;Bewaar D0
            moveq       #T15CharIn,d0        ;Vraag 1 keyboard/RS232 teken op
            trap        #15
            move.l      (sp)+,d0             ;Red D0
            rts

            
; --- CharOutT15 --------------------------------------------------------------
; 09-01-2022 Output 1 teken in D1 via trap-15 (user-mode)
CharOutT15: move.l      d0,-(sp)             ;Bewaar D0
            moveq       #T15CharOut,d0       ;Schrijf 1 teken naar RS232
            trap        #15
            move.l      (sp)+,d0             ;Red D0
            rts

            
; --- TRAP #1 ----------------------------------------------------------------
;1991: Trap #1: wordt 'super'
Trap1Super:    addq.l      #2,sp             ;Haal SR van stack
               rts                           ;Terug naar de userprogramma!

            
; --- TRAP #12 FUZIX ---------------------------------------------------------
;04-02-2022 Aanroep Fuzix
Trap12Fuzix:   move.l      FuzixSysCall,-(sp);Zet Fuzix Syscall adres op stack
               rts                           ;Start FUZIX systemcall!
            
        
; ============================================================================
; 1991 $0C Lees string (A0) met maximale lengte van D0.W
; Levert lengte af in D0.W, null terminated.
; CR einde String, BS ook gesuppord. ***SU-aanroep***
; Keep this routine as fast as possible to process S-record @115200 / 8MHz
ReadStr:movem.l d3/d4,-(SP)        ;Bewaar registers
        clr.l   d3                 ;D3.W = lengte huidige string
        MOVE.W  D0,D4              ;D4.W = max Lengte string
        SUBQ.W  #1,D4              ;Max lengte 1 kleiner ivm null-terminated
ReadS1: bsr     CharIn             ;Lees één teken
        cmpi.b  #' ',D0            ;Test op normaal teken
        blt.s   ReadS2             ;Afhandeling speciale tekens

        cmpi.b  #del,D0            ;LET OP: toets backspace = #127 (del)
        beq.s   ReadS4

        CMP.W   D4,D3              ;Maximale lengte ?
        bge.s   ReadS1             ;String vol, dan niets doen

        move.b  D0,(A0,D3.W)       ;Zet teken in de string
        addq.l  #1,D3              ;Hoog de lengte met één op
        tst.b   EchoInput          ;Echo teken?
        beq.s   ReadS1
        bsr     CharOut            ;Zet het teken op het scherm
        bra.s   ReadS1


ReadS2: cmpi.b  #lf,d0             ;CR of LF, idem (!Kea ASCII S-record upload)
        beq.s   ReadS5
        cmpi.b  #cr,D0             ;Afhandeling speciale tekens
        beq.s   ReadS5             ;CR? Dan klaar
        cmpi.b  #bs,D0             
        bne.s   ReadS1

ReadS4: tst.w   D3                 ;Afhandeling BS; lengte>0 ?
        beq.s   ReadS1             ;Als lengte nul is, dan niets doen
        subq.l  #1,D3              ;Verkort de lengte vd string met 1
        tst.b   EchoInput          ;Echo bs?
        beq.s   ReadS1

        move.l  a0,-(SP)           ;Bewaar register A0
        lea     BackspaceTxt(pc),a0
        bsr     StrOut             ;Print <Backspace><spatie><Backspace>
        move.l  (sp)+,a0           ;Haal register A0 terug
        bra.s   ReadS1

ReadS5: clr.b   0(a0,d3.w)         ;NULL-terminated string
        tst.b   EchoInput          ;Echo CrLf?
        beq.s   ReadS6
        bsr     T15CrLf            ;CR=einde van de input; print CR & LF
ReadS6: move.l  D3,D0              ;D0.l wordt huidige lengte
        movem.l (SP)+,d3/d4        ;Herstel registers
        rts


		include 	"bios_trap15.inc"
    
            
;--- Emergency  operation/geen RAM -------------------------------------------
;20-09-2023 Print sterretje en knipper LED; UART is al geïnitieerd
;Via menu dan eerst super ivm aansturen ACIA
MenuE:      trap        #1                      ;Wordt super om ACIA te bedienen
Emergency:  
;Initieer UART 16550 (MC68EC020/Nano68K) op 38.4 kBaud
            if UART16550
            lea.l       UART16550,A0
            move.b      #%00001101,FCR(A0)      ; enable FIFO
            move.b      #%10000011,LCR(A0)      ; 8 data bits, no parity, 1 stop bit, DLAB=1
            move.b      #BAUD_DIV_L,DLL(A0)     ; set divisor latch low byte
            move.b      #BAUD_DIV_U,DLM(A0)     ; set divisor latch high byte
            bclr.b      #7,LCR(A0)              ; disable divisor latch
            clr.b       SCR(A0)                 ; clear the scratch register
            move.b      #(1<<MCR_COPI),MCR(a0)  ; SPI COPI idles low

            bset.b      #MCR_LED,MCR(A0)        ;Led aan
            endif

            if System68K_Classic
            lea.l       AciaData,A0             ;A0 verwijst naar ACIA 8250
            endif

Emergency1: lea         EmergencyTxt(pc),a1     ;A1 wijst naar de tekst
            if UART16550                        ;LED-knipper einde string
            bchg.b      #MCR_LED,MCR(A0)        ;Toggle LED (on/off)
            else
            bchg.b      #PB2_LED,ViaPB          ;Toggle LED (on/off) 
            endif
Emergency2: move.b      (a1)+,d0                ;Zet teken in D0
            beq.s       Emergency1              ;Einde string, dan opnieuw laden           
            move.b      d0,(A0)                 ;Zend teken naar UART/ACIA
            move.l      #-1,d0                  ;Wachtlus
Emergency3: dbf         d0,Emergency3
            bra.s       Emergency2
 
            
;--- BIOS teksten ------------------------------------------------------------
;<27>[32m is groene tekst (voor FLterm); wordt direct gereset!?!
                  if System68K_Classic
InitTxt:          dc.b     esc,'[H',esc,'[2J',esc,'[1m68K/Classic ',esc,'[0m'
                  dc.b     ' (MC68HC000/16MHz) by Steven (july 1991).',cr,lf
                  dc.b     'Hardware revision RTC/SD-card (july 2021).',cr,lf
                  endif;
                  
                  if System68K_Nano
InitTxt:          dc.b     esc,'[H',esc,'[2J',esc,'[1m68K/Nano ',esc,'[0m'
                  dc.b     '(MC68HC000/24MHz) hardware design by Matt Sarnoff (2020), build 2023.',cr,lf
                  endif;

                  if SystemEC020SBC
InitTxt:          dc.b     esc,'[H',esc,'[2J',esc,'[1m ByteBlazer020 (MC68EC020/32Mhz)',esc,'[0m'
                  dc.b     ', by Steven (october 2023).',cr,lf
                  endif;

                  if System68K_Thunderbird
InitTxt:          dc.b     esc,'[H',esc,'[2J',esc,'[1m68K/Thunderbird ',esc,'[0m'
                  dc.b     '(MC68HC000-FN20/32MHz) by Steven (august 2024).',cr,lf
                  endif;
                  
                  if BIOSinRAM                  ;Extra melding indien in RAM
                  dc.b     esc,'[1m***BIOS in RAM*** ',esc,'[0m'
                  endif;
                  
                  dc.b     'Multi board BIOS''91 by Steven (06-10-2024).',cr,lf,cr,lf
                  dc.b      esc,'[1mHardware check:',esc,'[0m',cr,lf,0
                  
VectorTxt:        dc.b     '-RAM:  Copy vectors to RAM ',0
VectorOkTxt:      dc.b     '(OK).',cr,lf,0
VectorFailedTxt:  dc.b     '(FAILED).',cr,lf,0

ROMcheckTxt:      dc.b     '-ROM:  ',0
ROMfreeTxt:       dc.b     ' kB (',0
ROMbytesTxt:      dc.b     ' bytes free).',cr,lf,0

RAMevenTxt:       dc.b     '-RAM:  Even bank ',0
RAMevenOkTxt:     dc.b     '(OK), odd bank ',0
RAMevenErrorTxt:  dc.b     '(FAILED), odd bank ',0
RAModdOkTxt:      dc.b     '(OK).',cr,lf,0
RAModdErrorTxt:   dc.b     '(FAILED).',cr,lf,0

MEMtxt:           dc.b     '         0 kB RAM.',cr,0;
MEMbsTxt:         dc.b     cr, '     ',0       ;Terug naar begin vd regel

RTCtxt:           dc.b     cr,lf,'-VIA:  RTC DS1307: 1 sec tick (CA1) ',0  
RTCokTxt:         dc.b     '(OK).',cr,lf,0
RTCerrorTxt:      dc.b     '(FAILED) no RTC.',cr,lf,0

TimerTxt:         dc.b     '-CPU:  ',0
TimerErrorTxt:    dc.b     '(FAILED) (timer-1, RTC, INT 0x4 or RAM)',cr,lf,0
TimerOkTxt:       dc.b     '(OK).',cr,lf,0
TimerCPUtxt       dc.b     '000 Hz.',cr,lf,0

OscillatorTxt:    dc.b     cr,lf,'-UART: 100 Hz oscillator ',0
OscillatorOkTxt:  dc.b     '(OK)',0
OscillatorErrorTxt:dc.b    '(FAILED)',0

CPUclock1Txt:     dc.b     cr,lf,'-CPU:  Systemclock approx.',0
CPUclock2Txt:     dc.b     ' MHz',0

UARTtxt:          dc.b     cr,lf,'-SYS:  Enable RAM ',0
UARTtxt1:         dc.b     '(OK), enable instruction cache ',0
UARTtxt2:         dc.b     '(OK), enable ISR ',0
UARTokTxt:        dc.b     '(OK). System ready to go!',cr,lf,0 

MenuTxt:          dc.b     Xon,cr,lf, esc,'[1m'
                  dc.b     'Program:                            Test & exceptions:',esc,'[0m',cr,lf
                  dc.b     ' B) Boot system                      A)'
                  
                  if SystemEC020SBC
                  dc.b     ' Alternate cache status',cr,lf
                  else
                  dc.b     ' Address error',cr,lf
                  endif;
                  
                  dc.b     ' H) Hexdump                          C) CHK-trap',cr,lf
                  dc.b     ' L) Load & run <hi><lo><data><xor>   D) Division by zero',cr,lf
                  dc.b     ' R) Run program                      E) Emergency operation',cr,lf
                  dc.b     ' S) S-record load & run              I) Illegal instruction',cr,lf
                  dc.b     ' X) X-modem load & run (NOT WORKING) ',cr,lf
                  dc.b     cr,lf,esc,'[1m'
                  dc.b     'Benchmarks:                         Setup:',esc,'[0m',cr,lf
                  dc.b     ' F) Fibonacci+Sieve of Eratosthenes  T) Read/set RTC',cr,lf
                  dc.b     ' P) Primes                           M) Memory map',cr,lf,0
CommandTxt:       dc.b     cr,lf,esc,'[1mCommand',esc,'[0m (?=help): ',0


AbTxt:            dc.b     Xon,lf,lf,lf,lf,lf
                  dc.b     esc,'[19;1H    -->',esc,'[1m Access to ',0,' caused a',0
AdrTxt:           dc.b     'n address error',0
BusTxt:           dc.b     ' bus error',0
InstTxt:          dc.b     '    -->',esc,'[1m Illegal instruction',0
DivTxt:           dc.b     '    -->',esc,'[1m Division by zero',0
ChkTxt:           dc.b     '    -->',esc,'[1m Rangecheck (CHK) error',0
TrapVTxt:         dc.b     '    -->',esc,'[1m Overflow error',0
PrivTxt:          dc.b     '    -->',esc,'[1m Privilege violation',0
TraceTxt:         dc.b     '    -->',esc,'[1m Trace vector',0
SpurTxt:          dc.b     '    -->',esc,'[1m Spurious error',0
ResetTxt:         dc.b     '    -->',esc,'[1m Reset command',0

Err0Txt:          dc.b     Xon,lf,lf,lf,lf,lf,esc,'[19;1H',0
ErrTxt:           dc.b     ' at PC: ',0,' (=',0

RegTxt:           dc.b     ')',esc,'[0m',esc,'[K',esc,'[0m',esc,'[19;71H<--',cr,lf
                  dc.b     '  | D0 = ',0,'     D1 = ',0,'     D2 = ',0,'     D3 = ',0,' |',esc,'[K',cr,lf
                  dc.b     '  | D4 = ',0,'     D5 = ',0,'     D6 = ',0,'     D7 = ',0,' |',esc,'[K',cr,lf
                  dc.b     '  | A0 = ',0,'     A1 = ',0,'     A2 = ',0,'     A3 = ',0,' |',esc,'[K',cr,lf
                  dc.b     '  | A4 = ',0,'     A5 = ',0,'     A6 = ',0,'     SP = ',0,' |',esc,'[K',0

LaadTxt:          dc.b     cr,lf,'Loading...',0
LaadOkTxt:        dc.b     ' Start program in RAM...',cr,lf,0
LaadErrTxt:       dc.b     cr,lf,'checksum error.',cr,lf,0
EindPrTxt:        dc.b     cr,lf,esc,'[24;1H',esc,'[1mProgram terminated',esc,'[0m'
                  dc.b     '. Press any key to continue...',esc,'[K',0           
Trap15Txt         dc.b     'Trap #15 function not yet implemented: 0x',0
CrLfTxt           dc.b     cr,lf,0
BackspaceTxt      dc.b     bs,' ',bs,0
TimeShowTxt       dc.b     cr,lf,'Current time:        ',0
TimeSetTxt        dc.b     cr,lf,'Set time (hh:mm:ss): ',0

Fib16Txt:         dc.b     esc,'[H',esc,'[2J',esc,'[1mBenchmarks',esc,'[0m',cr,lf
                  dc.b     '1.   10x Fibonacci16(24) = ',0
Fib32Txt:         dc.b     '2.   10x Fibonacci32(24) = ',0
SieveTxt:         dc.b     '3.  100x Sieve of Eratosthenes, s=8190, '
                  dc.b     'primes found: ',0
RamCopyByteTxt:   dc.b     '4. 1000x Copy 4kB from EEPROM to RAM (B)',0
RamCopyWordTxt:   dc.b     '5. 1000x Copy 4kB from EEPROM to RAM (W)',0
RamCopyLongTxt:   dc.b     '6. 1000x Copy 4kB from EEPROM to RAM (L)',0
BitToggleTxt:     dc.b     '7. IO bit toggle, freq(kHz)=',0
BenchmarkReadyTxt:dc.b     '. Ready in ',0
BenchmarkMsTxt:   dc.b     ' ms.',cr,lf,0


HexdumpAddrTxt:   dc.b     cr,lf,'Address (or Stop): ',0
HexdumpColonTxt:  dc.b     ': ',0


SRecStartTxt      dc.b     cr,lf,esc,'[1mLoad S-records: ',esc,'[0m',cr,lf,0
SChkFoutTxt:      dc.b     cr,lf,'Wrong checkum in line ',0
SRecFoutTxt:      dc.b     cr,lf,'Wrong format in line ',0
SRecRunTxt:       dc.b     cr,lf,'Start program (user space) at 0x',0
SRecKbTxt:        dc.b     ' kB',cr,0
SPrgErrorTxt:     dc.b     'No program loaded!',cr,lf,0

Leap1Txt:         dc.b     ' bytes loaded in ',0
Leap2Txt:         dc.b     ' sec (',0
Leap3Txt:         dc.b     ' bytes/sec)',cr,lf,0

XModemStartTxt:   dc.b     cr,lf,'XModem start: ',0
XModemNAKtxt:     dc.b     NAK,0
XModemACKtxt:     dc.b     ACK,0
XModemCANtxt:     dc.b     CAN,CAN,0
XModemCancelTxt:  dc.b     'XModem canceled by sender',cr,lf,0
XModemTimeoutTxt: dc.b     'XModem timeout (no SOH)',cr,lf,0
XModemCRCtxt:     dc.b     'XModem CRC error',cr,lf,0
XModemSRECtxt:    dc.b     'XModem SREC error',cr,lf,0
XModemBloknrfoutTxt:
                  dc.b     CAN,CAN,'XModem bloknumber error',cr,lf,0
XModemKlaarTxt:   dc.b     ACK,cr,lf,'XModem transfer completed!',cr,lf,0

EmergencyTxt:     dc.b     cr,lf,'*Emergency program*',0

CacheEnableTxt:   dc.b     cr,lf,'CPU instruction cache enabled',cr,lf,0
CacheDisableTxt:  dc.b     cr,lf,'CPU instruction cache disabled',cr,lf,0


                  if System68K_Classic
;20-05-2024 Helptext memorymap (68000SBC / the original):
MemoryMapTxt:     dc.b     esc,'[H',esc,'[2J',esc,'[1mMemory map Classic (MC68HC000):',esc,'[0m',cr,lf
                  dc.b     '0x0.0000 - 0x0.1FFF: ROM  AT28C64B (2x8 kB)',cr,lf
                  dc.b     '0x1.0000 - 0x2.FFFF: RAM  62256 (4x32 kB)',cr,lf
                  dc.b     '0x3.0000 - 0x3.0007: VIA  6522',cr,lf
                  dc.b     '0x3.0001 - 0x3.0007: UART 6850',cr,lf,0
                  endif

                  if System68K_Nano
;Helptext memory map (68K nano):
MemoryMapTxt:     dc.b     esc,'[H',esc,'[2J',esc,'[1mMemory map 68K-nano (MC68HC000):',esc,'[0m',cr,lf
                  dc.b     '0x00.0000 - 0x00.2FFF: ROM  AT28C64B (2x  8 kB)',cr,lf
                  dc.b     '0xC0.0000 - 0xCF.FFFF: RAM  AS6C4008 (2x512 kB)',cr,lf
                  dc.b     '0x90.0000 - 0x90.000F: IDE/CF (SD/card)',cr,lf
                  dc.b     '0xA0.0000 - 0xA0.000F: UART 16550',cr,lf,0
                  endif

                  if SystemEC020SBC
;Helptext memory map (68EC020):
MemoryMapTxt:     dc.b     esc,'[H',esc,'[2J',esc,'[1mMemory map ByteBlaze020 (MC68EC020):',esc,'[0m',cr,lf
                  dc.b     '0x00.0000 - 0x00.FFFF: ROM  W27C514 (64 kB)',cr,lf
                  dc.b     '0x08.0000 - 0x0F.FFFF: RAM  AS6C4008 (512 kB)',cr,lf
                  dc.b     '0x10.0000 - 0x10.0007: IDE/CF (SD/card)',cr,lf
                  dc.b     '0x18.0000 - 0x18.0007: UART 16550',cr,lf,0
                  endif

                  if System68K_Thunderbird
;Helptext memory map (MC68HC0000FN20):
MemoryMapTxt:     dc.b     esc,'[H',esc,'[2J',esc,'[1mMemory map Thunderbird (MC6HC000-FN20):',esc,'[0m',cr,lf
                  dc.b     '0x00.0000 - 0x0F.FFFF: RAM  AS6C4008-55 (2x512 kB)',cr,lf
                  dc.b     '0x10.0000 - 0x10.0007: UART 16550',cr,lf
                  dc.b     '0x20.0000 - 0x21.FFFF: ROM  W27C514-45  (2x64 kB)',cr,lf
                  dc.b     '0x30.0000 - 0x30.0007: IDE/CF (SD-card)',cr,lf
                  endif

                  align 2                 ;Align 32 bits



;--- EINDE BIOS --------------------------------------------------------------      
            align    2              ;Maak einde 32 bits; nodig voor RAM-test
ProgEnd:                            ;Start memory geheugen dat getest kan worden
            END